
# def pretty_freqs():
#     ["KHz", "MHz", "GHz", "THz"]
#     {:.2f}"\
#                 .format(fft_bin_resolution)

# In the electrical domain,

# https://inst.eecs.berkeley.edu/~ee232/sp17/discussions/Discussion%207%20-%20Photonic%20circuit%20simulation.pptx

# in the optical domain, the impedance would just be that of free space?

from math import ceil
from .backend import backend as bd
from .detectors import LineDetector, BlockDetector, CurrentDetector
from .sources import SoftArbitraryPointSource

class FrequencyRoutines:
    # originally called FrequencyDomain, renamed because 'fd' is already a namespace

    #inputs:
    # One of, or a list of:
    # - a single plain array or numpy array of values (like the waveform_array input to SoftArb)
    # only works with FFT(), not impedance.
    # The length does not need to correspond to the simulation length (e.g. it can be shorter);
    # The timestep is assumed to be the same, however.
    #
    # - a single plain detector, of any type,
    #
    #
    # - a single SoftArbitraryPointSource object,
    # works with impedance or 1-port s-param,
    # - an array of SoftArbitraryPointSource
    # for multi-port S-param,
    # - the tuple (detector, current_detector) or [(detector, current_detector),...]
    # (the impedance is then assumed to be that of free-space).
    # FIXME:
    # free-space impedance should also eventually
    # depend on the permittivity and permeability of the detector cell
    #
    # converts it into the frequency domain.
    # frequency in Hz, not angular frequency



    def __init__(self, grid, objs):
        self.grid = grid
        self.objs = objs



    def compute_padding(self, input_data, dt, freq_window_tuple=None, fft_num_bins_in_window=None,
                                            fft_bin_freq_resolution=None):
        '''
        input_data must be a 1d array with the time history of one detector. the length of input_data
        is expected to be the same as time_steps_passed.

        Does not apply the padding to input_data,
        since multiple vectors will probably need the same padding and it
        seems to make more sense to do that in the caller.

        - fft_num_bins
        - FFT bin resolution (optional) Hz.
        '''

        # This seems like magic - how could this be?

        # the key is that indeed no extra information is being added;
        # sinc interplolation

        # https://dsp.stackexchange.com/questions/31783/mathematical-justification-for-zero-padding
        # https://dsp.stackexchange.com/questions/741/why-should-i-zero-pad-a-signal-before-taking-the-fourier-transform
        # https://dsp.stackexchange.com/questions/24410/zero-padding-of-fft/24426#24426

        # Paraphrasing from the great explanation from
        # https://www.bitweenie.com/listings/fft-zero-padding/
        # there are two different frequency resolutions at play.
        # The first one, a physical resolution,

        # there are other ways to get a higher bin resolution without explicitly padding, too


        input_length = input_data.shape[0]

        if(freq_window_tuple == None):
            begin_freq = 0
            # from numpy fftfreq docs
            end_freq = (input_length/2.0) / (dt*input_length) # off by one?
        else:
            begin_freq, end_freq = freq_window_tuple

        end_time = input_length * dt

        if(fft_num_bins_in_window == None and not fft_bin_freq_resolution == None):
            fft_num_bins_in_window = ((begin_freq-end_freq)/fft_bin_freq_resolution)

        elif(not (fft_num_bins_in_window or fft_bin_freq_resolution)):
            # the window is the default, whole array (even if it's trimmed later)
            fft_num_bins_in_window = input_data.shape[0]

        waveform_frequency_resolution = 1.0 / end_time
        fft_bin_resolution = (1.0 / dt) / (fft_num_bins_in_window)
        print("Waveform data has an intrinsic resolution of: {:.2E} Hz"\
                    .format(waveform_frequency_resolution))
        print("FFT bin: {:.2E} Hz"\
                    .format(fft_bin_resolution))

        required_padding = ceil(fft_num_bins_in_window / ((end_freq-begin_freq) * dt)) - input_length



        return required_padding, end_time


        # assumes a uniform timestep. It might be useful to add a .times vector to the grid
        # if the timestep is made variable at some point.
        # on the other hand, doing a non-uniform FFT is probably non-trivial at this point anyway
        # times = bd.linspace(0,end_time + (required_padding*dt),
        #                                         (input_length+required_padding))


    def compute_frequencies(self,length_with_padding, dt, freq_window_tuple=None):
        '''
        Outputs the
        The indexes are for the real-frequency part of the span.
        '''

        spectrum_freqs = bd.fftfreq(length_with_padding, d=dt)

        if(freq_window_tuple == None):
            begin_freq_idx = 0
            # end_freq_idx = ((length_with_padding/2)-1)
            end_freq_idx = -1
        else:
            #closest frequencies
            begin_freq_idx = bd.abs(spectrum_freqs - begin_freq).argmin()
            end_freq_idx = bd.abs(spectrum_freqs - end_freq).argmin()

        return spectrum_freqs, begin_freq_idx, end_freq_idx

    # UNTESTED
    def S_parameters(waveform, node):
        raise NotImplementedError
        # outputs the S params for each frequency

        # [1]Kurokawa K.
        # Power Waves and the Scattering Matrix.
        # IEEE Transactions on Microwave Theory and Techniques 1965
        # ;13:194–202. https://doi.org/10.1109/TMTT.1965.1125964.

        # Need to construct the matrix from Kurokawa "power wave" coefficients:
        # a_1 =  #incident
        # b_1 =  #reflected
        # Wikipedia has a great section:
        # https://en.wikipedia.org/wiki/Scattering_parameters

        # inverse matrix?
        #
        # if(isinstance(self.objs, list)):
        #
        # elif(isinstance(self.objs,SoftArbitraryPointSource)):
        # elif(isinstance(self.objs,tuple) and isinstance(self.objs[0],(BlockDetector,LineDetector))
        #                                 and isinstance(self.objs[1],CurrentDetector)):
        # else:
        #     raise ValueError("Sorry, FFT can't yet interpret the argument given.")

    def export_touchstone_s2p():
        raise NotImplementedError



    def FFT(self, freq_window_tuple=None, fft_num_bins_in_window=None,
                                            fft_bin_freq_resolution=None):
        '''
        FIXME: If is type BlockDetector or other similar output,
        for now the fft of only one index is returned.
        '''
        #scaling factor here?

        if(self.grid.time_steps_passed == 0):
            return [], []
        #I now see how Multiple Dispatch might be effective at speeding things up
        if(bd.is_array(self.objs)):
            input_data = self.objs
        elif(isinstance(self.objs,(LineDetector, BlockDetector))):
            #FIXME:
            input_data = self.objs.E[:][0][0][0]
        elif(isinstance(self.objs,SoftArbitraryPointSource)):
            #FIXME:
            input_data = self.objs.source_voltage[:][0][0][0]
        elif(isinstance(self.objs,CurrentDetector)):
            #FIXME:
            input_data = self.objs.I[:][0][0][0]
        else:
            raise ValueError("Sorry, FFT can't yet interpret the argument given.")

        input_data = bd.array(input_data)
        required_padding, _ = self.compute_padding(input_data,
                                                self.grid.time_step,
                                                freq_window_tuple=freq_window_tuple,
                                                fft_num_bins_in_window=fft_num_bins_in_window,
                                                fft_bin_freq_resolution=fft_bin_freq_resolution)

        input_data = bd.pad(input_data, (0, required_padding), 'edge')

        spectrum = bd.fft(input_data)

        spectrum_freqs, begin_freq_idx, end_freq_idx = self.compute_frequencies(
                                        input_data.shape[0], self.grid.time_step,
                                         freq_window_tuple=freq_window_tuple)

        return spectrum_freqs[begin_freq_idx:end_freq_idx], spectrum[begin_freq_idx:end_freq_idx]



    def impedance(self, freq_window_tuple=None, fft_num_bins_in_window=None,
                                            fft_bin_freq_resolution=None):
        '''
        Takes a single "node", a current and voltage record. The form of:
        SoftArbitraryPointSource or tuple.
        '''
        if(self.grid.time_steps_passed == 0):
            return [], []

        if(isinstance(self.objs, SoftArbitraryPointSource)):
            voltage_array = bd.array(self.objs.source_voltage)
            current_array = bd.array(self.objs.current_detector.I) #FIXME: this definitely isn't right

        elif(isinstance(self.objs,tuple) and isinstance(self.objs[0],(BlockDetector,LineDetector))
                                        and isinstance(self.objs[1],CurrentDetector)):
            #FIXME:
            voltage_array = bd.array(self.objs[0].E)[:][0,0,0]
            current_array = bd.array(self.objs[1].I)
        else:
            raise ValueError("Sorry, FFT can't yet interpret the argument given.")

        required_padding, _ = self.compute_padding(voltage_array,
                                                self.grid.time_step,
                                                freq_window_tuple=freq_window_tuple,
                                                fft_num_bins_in_window=fft_num_bins_in_window,
                                                fft_bin_freq_resolution=fft_bin_freq_resolution)

        voltage_array = bd.reshape(voltage_array, self.grid.time_steps_passed) # flatten
        current_array = bd.reshape(current_array, self.grid.time_steps_passed)

        voltage_array = bd.pad(voltage_array, (0, required_padding), 'edge')
        current_array = bd.pad(current_array, (0, required_padding), 'edge')

        voltage_spectrum = bd.fft(voltage_array)
        current_spectrum = bd.fft(current_array)

        impedance_spectrum = bd.divide(voltage_spectrum, current_spectrum, out=
                                    bd.zeros_like(voltage_spectrum), where=current_spectrum!=0)


        spectrum_freqs, begin_freq_idx, end_freq_idx = self.compute_frequencies(
                                        voltage_array.shape[0], self.grid.time_step,
                                         freq_window_tuple=freq_window_tuple)

        return spectrum_freqs[begin_freq_idx:end_freq_idx], impedance_spectrum[begin_freq_idx:end_freq_idx]


        #
        # voltages = bd.pad(voltages, (0, required_length), 'edge')
        # currents = bd.pad(currents, (0, required_length), 'edge')
        #
        # voltage_spectrum = bd.fft(voltages)
        # current_spectrum = bd.fft(currents)
        #
        # spectrum_freqs, begin_freq_idx, end_freq_idx = \
        #                         compute_frequencies(length_with_padding, dt, freq_window_tuple=None)
        #
        # spectrum_freqs[begin_freq:end_freq]


    def plot_impedance(self, freq_window_tuple=None, fft_num_bins_in_window=None,
                                            fft_bin_freq_resolution=None):
        '''
        Frequencies are in Hz.
        '''

        plt.plot(times_padded, voltages)

        plt.plot(times_padded, currents)
        # plt.plot(, abs(voltage_spectrum[begin_freq:end_freq]), label="volt")
        plt.plot(spectrum_freqs[begin_freq:end_freq], abs(current_spectrum[begin_freq:end_freq]), label="curr")


        plt.figure()
        #
        # Z0 = scipy.constants.physical_constants['characteristic impedance of vacuum'][0]

        impedance_spectrum = abs(voltage_spectrum/current_spectrum)

        plt.plot(spectrum_freqs[begin_freq:end_freq],impedance_spectrum[begin_freq:end_freq])
        # plt.plot(spectrum_freqs[begin_freq:end_freq],impedance_spectrum[begin_freq:end_freq])
        plt.savefig('/tmp/impedance_spectrum.svg')
        # # plt.plot(spectrum_freqs,(voltage_spectrum/current_spectrum))
        # plt.plot(spectrum_freqs[begin_freq:end_freq],(voltage_spectrum[begin_freq:end_freq]/current_spectrum[begin_freq:end_freq]).real)
        # plt.plot(spectrum_freqs[begin_freq:end_freq],(voltage_spectrum[begin_freq:end_freq]/current_spectrum[begin_freq:end_freq]).imag)

        plt.show()
        # power_spectrum = -1.0*((voltage_spectrum[begin_freq:end_freq]*np.conj(current_spectrum[begin_freq:end_freq])).real)
        # power_spectrum /= np.linalg.norm(power_spectrum)
        # plt.plot(spectrum_freqs[begin_freq:end_freq],power_spectrum)



""" often used constants go here """

from math import pi

c: float = 299792458.0
""" speed of light """

X: int = 0
""" x-index (useful for indexing) """

Y: int = 1
""" y-index (useful for indexing) """

Z: int = 2
""" z-index (useful for indexing) """

mu0: float = 4e-7 * pi
""" vacuum permeability """

eps0: float = 1.0 / (mu0 * c ** 2)
""" vacuum permittivity """

eta0: float = mu0 * c
""" free space impedance """



""" common types """

## Standard Library Types
from typing import Union, Tuple, List
from numbers import Number

## Numpy types
from numpy import ndarray

## Torch types
try:
    from torch import Tensor
except ImportError:
    Tensor = None

## Backend types
if Tensor is not None:
    Tensorlike = Union[ndarray, Tensor]
else:
    Tensorlike = ndarray

## Composite types
ListOrSlice = Union[List[int], slice]
IntOrSlice = Union[int, slice]



""" Sources are objects that inject the fields into the grid.

Available sources:

- PointSource
- LineSource

"""
## Imports

# other
from math import pi, sin

# typing
from .typing_ import Tuple, Number, ListOrSlice, List
from numpy import ndarray

# relatvie
from .grid import Grid
from .backend import backend as bd
from .waveforms import *
from .detectors import CurrentDetector

## PointSource class
class PointSource:
    """A source placed at a single point (grid cell) in the grid"""

    def __init__(
        self,
        period: Number = 15,
        amplitude: float = 1.0,
        phase_shift: float = 0.0,
        name: str = None,
        pulse: bool = False,
        cycle: int = 5,
        hanning_dt: float = 10.0,
    ):
        """Create a LineSource with a gaussian profile

        Args:
            period: The period of the source. The period can be specified
                as integer [timesteps] or as float [seconds]
            amplitude: The electric field amplitude in simulation units
            phase_shift: The phase offset of the source.
            name: name of the source.
            pulse: Set True to use a Hanning window pulse instead of continuous wavefunction.
            cycle: cycles for Hanning window pulse.
            hanning_dt: timestep used for Hanning window pulse width (optional).

        """
        self.grid = None
        self.period = period
        self.amplitude = amplitude
        self.phase_shift = phase_shift
        self.name = name
        self.pulse = pulse
        self.cycle = cycle
        self.frequency = 1.0 / period
        self.hanning_dt = hanning_dt if hanning_dt is not None else 0.5 / self.frequency

    def _register_grid(self, grid: Grid, x: Number, y: Number, z: Number):
        """Register a grid for the source.

        Args:
            grid: the grid to place the source into.
            x: The x-location of the source in the grid
            y: The y-location of the source in the grid
            z: The z-location of the source in the grid

        Note:
            As its name suggests, this source is a POINT source.
            Hence it should be placed at a single coordinate tuple
            int the grid.
        """
        self.grid = grid
        self.grid.sources.append(self)
        if self.name is not None:
            if not hasattr(grid, self.name):
                setattr(grid, self.name, self)
            else:
                raise ValueError(
                    f"The grid already has an attribute with name {self.name}"
                )

        try:
            (x,), (y,), (z,) = x, y, z
        except (TypeError, ValueError):
            raise ValueError("a point source should be placed on a single grid cell.")
        self.x, self.y, self.z = grid._handle_tuple((x, y, z))
        self.period = grid._handle_time(self.period)

    def update_E(self):
        """Add the source to the electric field"""
        q = self.grid.time_steps_passed
        # if pulse
        if self.pulse:
            t1 = int(2 * pi / (self.frequency * self.hanning_dt / self.cycle))
            if q < t1:
                src = self.amplitude * hanning(
                    self.frequency, q * self.hanning_dt, self.cycle
                )
            else:
                # src = - self.grid.E[self.x, self.y, self.z, 2]
                src = 0
        # if not pulse
        else:
            src = self.amplitude * sin(2 * pi * q / self.period + self.phase_shift)
        self.grid.E[self.x, self.y, self.z, 2] += src

    def update_H(self):
        """Add the source to the magnetic field"""

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(period={self.period}, "
            f"amplitude={self.amplitude}, phase_shift={self.phase_shift}, "
            f"name={repr(self.name)})"
        )

    def __str__(self):
        s = "    " + repr(self) + "\n"
        x = f"{self.x}"
        y = f"{self.y}"
        z = f"{self.z}"
        s += f"        @ x={x}, y={y}, z={z}\n"
        return s


## LineSource class
class LineSource:
    """A source along a line in the FDTD grid"""

    def __init__(
        self,
        period: Number = 15,
        amplitude: float = 1.0,
        phase_shift: float = 0.0,
        name: str = None,
        pulse: bool = False,
        cycle: int = 5,
        hanning_dt: float = 10.0,
    ):
        """Create a LineSource with a gaussian profile

        Args:
            period: The period of the source. The period can be specified
                as integer [timesteps] or as float [seconds]
            amplitude: The amplitude of the source in simulation units
            phase_shift: The phase offset of the source.
            pulse: Set True to use a Hanning window pulse instead of continuous wavefunction.
            cycle: cycles for Hanning window pulse.
            hanning_dt: timestep used for Hanning window pulse width (optional).

        """
        self.grid = None
        self.period = period
        self.amplitude = amplitude
        self.phase_shift = phase_shift
        self.name = name
        self.pulse = pulse
        self.cycle = cycle
        self.frequency = 1.0 / period
        self.hanning_dt = hanning_dt if hanning_dt is not None else 0.5 / self.frequency

    def _register_grid(
        self, grid: Grid, x: ListOrSlice, y: ListOrSlice, z: ListOrSlice
    ):
        """Register a grid for the source.

        Args:
            grid: the grid to place the source into.
            x: The x-location of the source in the grid
            y: The y-location of the source in the grid
            z: The z-location of the source in the grid

        Note:
            As its name suggests, this source is a LINE source.
            Hence the source spans the diagonal of the cube
            defined by the slices in the grid.
        """
        self.grid = grid
        self.grid.sources.append(self)
        if self.name is not None:
            if not hasattr(grid, self.name):
                setattr(grid, self.name, self)
            else:
                raise ValueError(
                    f"The grid already has an attribute with name {self.name}"
                )

        self.x, self.y, self.z = self._handle_slices(x, y, z)

        self.period = grid._handle_time(self.period)

        L = len(self.x)
        vect = bd.array(
            (bd.array(self.x) - self.x[L // 2]) ** 2
            + (bd.array(self.y) - self.y[L // 2]) ** 2
            + (bd.array(self.z) - self.z[L // 2]) ** 2,
            bd.float,
        )

        self.profile = bd.exp(-(vect ** 2) / (2 * (0.5 * vect.max()) ** 2))
        self.profile /= self.profile.sum()
        self.profile *= self.amplitude

    def _handle_slices(
        self, x: ListOrSlice, y: ListOrSlice, z: ListOrSlice
    ) -> Tuple[List, List, List]:
        """Convert slices in the grid to lists

        This is necessary to make the source span the diagonal of the volume
        defined by the slices.

        Args:
            x: The x-location of the volume in the grid
            y: The y-location of the volume in the grid
            z: The z-location of the volume in the grid

        Returns:
            x, y, z: the x, y and z coordinates of the source as lists

        """

        # if list-indices were chosen:
        if isinstance(x, list) and isinstance(y, list) and isinstance(z, list):
            if len(x) != len(y) or len(y) != len(z) or len(z) != len(x):
                raise IndexError(
                    "sources require grid to be indexed with slices or equal length list-indices"
                )
            x = [self.grid._handle_distance(_x) for _x in x]
            y = [self.grid._handle_distance(_y) for _y in y]
            z = [self.grid._handle_distance(_z) for _z in z]
            return x, y, z

        # if a combination of list-indices and slices were chosen,
        # convert the list-indices to slices.
        # TODO: maybe issue a warning here?
        if isinstance(x, list):
            x = slice(
                self.grid._handle_distance(x[0]),
                self.grid._handle_distance(x[-1]),
                None,
            )
        if isinstance(y, list):
            y = slice(
                self.grid._handle_distance(y[0]),
                self.grid._handle_distance(y[-1]),
                None,
            )
        if isinstance(z, list):
            z = slice(
                self.grid._handle_distance(z[0]),
                self.grid._handle_distance(z[-1]),
                None,
            )

        # if we get here, we can assume slices:
        x0 = self.grid._handle_distance(x.start if x.start is not None else 0)
        y0 = self.grid._handle_distance(y.start if y.start is not None else 0)
        z0 = self.grid._handle_distance(z.start if z.start is not None else 0)
        x1 = self.grid._handle_distance(x.stop if x.stop is not None else self.grid.Nx)
        y1 = self.grid._handle_distance(y.stop if y.stop is not None else self.grid.Ny)
        z1 = self.grid._handle_distance(z.stop if z.stop is not None else self.grid.Nz)

        # we can now convert these coordinates into index lists
        m = max(abs(x1 - x0), abs(y1 - y0), abs(z1 - z0))
        if m < 2:
            raise ValueError("a LineSource should consist of at least two gridpoints")
        x = [v.item() for v in bd.array(bd.linspace(x0, x1, m, endpoint=False), bd.int)]
        y = [v.item() for v in bd.array(bd.linspace(y0, y1, m, endpoint=False), bd.int)]
        z = [v.item() for v in bd.array(bd.linspace(z0, z1, m, endpoint=False), bd.int)]

        return x, y, z

    def update_E(self):
        """Add the source to the electric field"""
        q = self.grid.time_steps_passed
        # if pulse
        if self.pulse:
            t1 = int(2 * pi / (self.frequency * self.hanning_dt / self.cycle))
            if q < t1:
                vect = self.profile * hanning(
                    self.frequency, q * self.hanning_dt, self.cycle
                )
            else:
                # src = - self.grid.E[self.x, self.y, self.z, 2]
                vect = self.profile * 0
        # if not pulse
        else:
            vect = self.profile * sin(2 * pi * q / self.period + self.phase_shift)
        # do not use list indexing here, as this is much slower especially for torch backend
        # DISABLED: self.grid.E[self.x, self.y, self.z, 2] = vect
        for x, y, z, value in zip(self.x, self.y, self.z, vect):
            self.grid.E[x, y, z, 2] += value

    def update_H(self):
        """Add the source to the magnetic field"""

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(period={self.period}, "
            f"amplitude={self.amplitude}, phase_shift={self.phase_shift}, "
            f"name={repr(self.name)})"
        )

    def __str__(self):
        s = "    " + repr(self) + "\n"
        x = f"[{self.x[0]}, ... , {self.x[-1]}]"
        y = f"[{self.y[0]}, ... , {self.y[-1]}]"
        z = f"[{self.z[0]}, ... , {self.z[-1]}]"
        s += f"        @ x={x}, y={y}, z={z}\n"
        return s


## PlaneSource class
class PlaneSource:
    """A source along a plane in the FDTD grid"""

    def __init__(
        self,
        period: Number = 15,
        amplitude: float = 1.0,
        phase_shift: float = 0.0,
        name: str = None,
        polarization: str = 'z',
    ):
        """Create a PlaneSource.

        Args:
            period: The period of the source. The period can be specified
                as integer [timesteps] or as float [seconds]
            amplitude: The amplitude of the source in simulation units
            phase_shift: The phase offset of the source.
            polarization: Axis of E-field polarization ('x','y',or 'z')
        """
        self.grid = None
        self.period = period
        self.amplitude = amplitude
        self.phase_shift = phase_shift
        self.name = name
        self.polarization = polarization

    def _register_grid(
        self, grid: Grid, x: ListOrSlice, y: ListOrSlice, z: ListOrSlice
    ):
        """Register a grid for the source.

        Args:
            grid: the grid to place the source into.
            x: The x-location of the source in the grid
            y: The y-location of the source in the grid
            z: The z-location of the source in the grid

        Note:
        As its name suggests, this source is a LINE source.
            Hence the source spans the diagonal of the cube
            defined by the slices in the grid.
        """
        self.grid = grid
        self.grid.sources.append(self)
        if self.name is not None:
            if not hasattr(grid, self.name):
                setattr(grid, self.name, self)
            else:
                raise ValueError(
                    f"The grid already has an attribute with name {self.name}"
                )

        self.x, self.y, self.z = self._handle_slices(x, y, z)

        self.period = grid._handle_time(self.period)

        x = bd.arange(self.x.start, self.x.stop, 1) - (self.x.start + self.x.stop) // 2
        y = bd.arange(self.y.start, self.y.stop, 1) - (self.y.start + self.y.stop) // 2
        z = bd.arange(self.z.start, self.z.stop, 1) - (self.z.start + self.z.stop) // 2
        xvec, yvec, zvec = bd.broadcast_arrays(
            x[:, None, None], y[None, :, None], z[None, None, :]
        )
        _xvec = bd.array(xvec, float)
        _yvec = bd.array(yvec, float)
        _zvec = bd.array(zvec, float)

        profile = bd.ones(_xvec.shape)
        self.profile = self.amplitude * profile

    def _handle_slices(
        self, x: ListOrSlice, y: ListOrSlice, z: ListOrSlice
    ) -> Tuple[List, List, List]:
        """Validate slices and calculate center of plane

        Args:
            x: The x-location of the volume in the grid
            y: The y-location of the volume in the grid
            z: The z-location of the volume in the grid

        Returns:
            x, y, z: the x, y and z coordinates of the source as slices

        """
        # ensure all slices
        if not isinstance(x, slice):
            if isinstance(x, list):
                (x,) = x
            x = slice(
                self.grid._handle_distance(x), self.grid._handle_distance(x) + 1, None
            )
        if not isinstance(y, slice):
            if isinstance(y, list):
                (y,) = y
            y = slice(
                self.grid._handle_distance(y), self.grid._handle_distance(y) + 1, None
            )
        if not isinstance(z, slice):
            if isinstance(z, list):
                (z,) = z
            z = slice(
                self.grid._handle_distance(z), self.grid._handle_distance(z) + 1, None
            )

        # if we get here, we can assume slices:
        x0 = self.grid._handle_distance(x.start if x.start is not None else 0)
        y0 = self.grid._handle_distance(y.start if y.start is not None else 0)
        z0 = self.grid._handle_distance(z.start if z.start is not None else 0)
        x1 = self.grid._handle_distance(x.stop if x.stop is not None else self.grid.Nx)
        y1 = self.grid._handle_distance(y.stop if y.stop is not None else self.grid.Ny)
        z1 = self.grid._handle_distance(z.stop if z.stop is not None else self.grid.Nz)

        # make sure all slices have a start, stop and no step:
        x = (
            slice(x0, x1)
            if x0 < x1
            else (slice(x1, x0) if x0 > x1 else slice(x0, x0 + 1))
        )
        y = (
            slice(y0, y1)
            if y0 < y1
            else (slice(y1, y0) if y0 > y1 else slice(y0, y0 + 1))
        )
        z = (
            slice(z0, z1)
            if z0 < z1
            else (slice(z1, z0) if z0 > z1 else slice(z0, z0 + 1))
        )

        if [x.stop - x.start, y.stop - y.start, z.stop - z.start].count(0) > 0:
            raise ValueError(
                "Given location for PlaneSource results in slices of length 0!"
            )
        if [x.stop - x.start, y.stop - y.start, z.stop - z.start].count(1) == 0:
            raise ValueError("Given location for PlaneSource is not a 2D plane!")
        if [x.stop - x.start, y.stop - y.start, z.stop - z.start].count(1) > 1:
            raise ValueError(
                "Given location for PlaneSource should have no more than one dimension in which it's flat.\n"
                "Use a LineSource for lower dimensional sources."
            )

        self._Epol = 'xyz'.index(self.polarization)
        if (x.stop - x.start == 1 and self.polarization == 'x') or \
           (y.stop - y.start == 1 and self.polarization == 'y') or \
           (z.stop - z.start == 1 and self.polarization == 'z'):
            raise ValueError(
                "PlaneSource cannot be polarized perpendicular to the orientation of the plane."
            )
        _Hpols = [(z,1,2), (z,0,2), (y,0,1)][self._Epol]
        if _Hpols[0].stop - _Hpols[0].start == 1:
            self._Hpol = _Hpols[1]
        else:
            self._Hpol = _Hpols[2]

        return x, y, z

    def update_E(self):
        """Add the source to the electric field"""
        q = self.grid.time_steps_passed
        vect = self.profile * sin(2 * pi * q / self.period + self.phase_shift)
        self.grid.E[self.x, self.y, self.z, self._Epol] = vect

    def update_H(self):
        """Add the source to the magnetic field"""
        q = self.grid.time_steps_passed
        vect = self.profile * sin(2 * pi * q / self.period + self.phase_shift)
        self.grid.H[self.x, self.y, self.z, self._Hpol] = vect

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(period={self.period}, "
            f"amplitude={self.amplitude}, phase_shift={self.phase_shift}, "
            f"name={repr(self.name)}, polarization={repr(self.polarization)})"
        )

    def __str__(self):
        s = "    " + repr(self) + "\n"
        x = f"[{self.x.start}, ... , {self.x.stop}]"
        y = f"[{self.y.start}, ... , {self.y.stop}]"
        z = f"[{self.z.start}, ... , {self.z.stop}]"
        s += f"        @ x={x}, y={y}, z={z}\n"
        return s


class SoftArbitraryPointSource:
    r"""

    A source placed at a single point (grid cell) in the grid.
    This source is special: it's both a source and a detector.

    Unlike the other sources, the input is a voltage, not an electric field.
    (really? why? should we convert back and forth?)

    For electrical measurements I've only needed a single-index source,
    so I don't know how the volume/line sources above work.
    We want the FFT function to operate over any detector.
    Maybe all sources should take an arbitary waveform argument?

    Each index in the *waveform* array represents 1 value at a timestep.

    There are many different *geometries* of "equivalent sources".
    The detector/source paradigm used in /fdtd might perhaps not correspond to this in an ideal
    fashion.

    It's not intuitively clear to me what a "soft" source would imply in the optical case, or what
    impedance even means for a laser.

    /fdtd/ seems to have found primary use in optical circles,
    so the default Z should probably be 0.

    "Whilst established for microwaves and electrical circuits,
    this concept has only very recently been observed in the optical domain,
    yet is not well defined or understood."[1]

    [1]: Optical impedance of metallic nano-structures, M. Mazilu and K. Dholakia
    https://doi.org/10.1364/OE.14.007709

    [2]: http://www.gwoptics.org/learn/02_Plane_waves/01_Fabry_Perot_cavity/02_Impedance_matched.php

    /\/\-
    """

    def __init__(
        self, waveform_array: ndarray, name: str = None, impedance: float = 0.0
    ):
        """Create



        Args:
            waveform_array
        """
        self.grid = None
        self.name = name
        self.current_detector = None
        self.waveform_array = waveform_array
        self.impedance = impedance
        self.input_voltage = []  # voltage hard-imposed by the source
        self.source_voltage = []  #
        # "field" rather than "voltage" might be more meaningful
        # FIXME: these voltage time histories have a different dimensionality

    def _register_grid(self, grid: Grid, x: Number, y: Number, z: Number):
        """Register a grid for the source.

        Args:
            grid: the grid to place the source into.
            x: The x-location of the source in the grid
            y: The y-location of the source in the grid
            z: The z-location of the source in the grid

        Note:
            As its name suggests, this source is a POINT source.
            Hence it should be placed at a single coordinate tuple
            int the grid.
        """
        self.grid = grid
        self.grid.sources.append(self)
        if self.name is not None:
            if not hasattr(grid, self.name):
                setattr(grid, self.name, self)
            else:
                raise ValueError(
                    f"The grid already has an attribute with name {self.name}"
                )

        try:
            (x,), (y,), (z,) = x, y, z
        except (TypeError, ValueError):
            raise ValueError("a point source should be placed on a single grid cell.")
        self.x, self.y, self.z = grid._handle_tuple((x, y, z))

        if self.name is not None:
            detector_name += "_I"
        else:
            detector_name = None

        self.current_detector = CurrentDetector(name=detector_name)
        grid[x, y, z] = self.current_detector

    def update_E(self):

        # It is important that this step happen between the E-field update and the
        # H-field update.

        if self.grid.time_steps_passed < self.waveform_array.shape[0]:
            # check for off-by-one error here
            input_voltage = self.waveform_array[self.grid.time_steps_passed]
        else:
            input_voltage = 0.0  # one could taper the last value off smoothly instead

        if self.grid.time_steps_passed > 0:
            current = self.current_detector.I[-1][0][0][0]
        else:
            current = 0.0

        if self.impedance > 0:
            source_resistive_voltage = self.impedance * current
            output_voltage = input_voltage + source_resistive_voltage
        else:
            output_voltage = input_voltage

        # right now, this does not compensate for the cell's permittivity!

        self.grid.E[self.x, self.y, self.z, 2] += (
            output_voltage / self.grid.grid_spacing
        )

        self.input_voltage.append([[[input_voltage]]])
        self.source_voltage.append([[[output_voltage]]])

    def update_H(self):
        pass

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def __str__(self):
        s = "    " + repr(self) + "\n"
        x = f"{self.x}"
        y = f"{self.y}"
        z = f"{self.z}"
        s += f"        @ x={x}, y={y}, z={z}\n"
        return s



""" The objects to place in the grid.

Objects define all the regions in the grid with a modified update equation,
such as for example regions with anisotropic permittivity etc.

Available Objects:
 - Object
 - AnisotropicObject

"""

## Imports

# typing
from .typing_ import Tensorlike, ListOrSlice

# relative
from .grid import Grid
from .backend import backend as bd
from . import constants as const


## Object
class Object:
    """ An object to place in the grid """

    def __init__(self, permittivity: Tensorlike, name: str = None):
        """
        Args:
            permittivity: permittivity tensor
            name: name of the object (will become available as attribute to the grid)
        """
        self.grid = None
        self.name = name
        self.permittivity = bd.array(permittivity)

    def _register_grid(
        self, grid: Grid, x: ListOrSlice, y: ListOrSlice, z: ListOrSlice
    ):
        """Register the object to the grid

        Args:
            grid: the grid to register the object into
            x: the x-location of the object in the grid
            y: the y-location of the object in the grid
            z: the z-location of the object in the grid
        """
        self.grid = grid
        self.grid.objects.append(self)
        if self.name is not None:
            if not hasattr(grid, self.name):
                setattr(grid, self.name, self)
            else:
                raise ValueError(
                    f"The grid already has an attribute with name {self.name}"
                )
        self.x = self._handle_slice(x, max_index=self.grid.Nx)
        self.y = self._handle_slice(y, max_index=self.grid.Ny)
        self.z = self._handle_slice(z, max_index=self.grid.Nz)

        self.Nx = abs(self.x.stop - self.x.start)
        self.Ny = abs(self.y.stop - self.y.start)
        self.Nz = abs(self.z.stop - self.z.start)

        # set the permittivity of the object
        if bd.is_array(self.permittivity) and len(self.permittivity.shape) == 3:
            self.permittivity = self.permittivity[:, :, :, None]
        self.inverse_permittivity = (
            bd.ones((self.Nx, self.Ny, self.Nz, 3)) / self.permittivity
        )

        # set the permittivity values of the object at its border to be equal
        # to the grid permittivity. This way, the object is made symmetric.
        if self.Nx > 1:
            self.inverse_permittivity[-1, :, :, 0] = self.grid.inverse_permittivity[
                -1, self.y, self.z, 0
            ]
        if self.Ny > 1:
            self.inverse_permittivity[:, -1, :, 1] = self.grid.inverse_permittivity[
                self.x, -1, self.z, 1
            ]
        if self.Nz > 1:
            self.inverse_permittivity[:, :, -1, 2] = self.grid.inverse_permittivity[
                self.x, self.y, -1, 2
            ]

        self.grid.inverse_permittivity[self.x, self.y, self.z] = 0

    def _handle_slice(self, s: ListOrSlice, max_index: int = None) -> slice:
        if isinstance(s, list):
            if len(s) == 1:
                return slice(s[0], s[0] + 1, None)
            raise IndexError(
                "One can only use slices or single indices to index the grid for an Object"
            )
        if isinstance(s, slice):
            start, stop, step = s.start, s.stop, s.step
            if step is not None and step != 1:
                raise IndexError(
                    "Can only use slices with unit step to index the grid for an Object"
                )
            if start is None:
                start = 0
            if start < 0:
                start = max_index + start
            if stop is None:
                stop = max_index
            if stop < 0:
                stop = max_index + stop
            return slice(start, stop, None)
        raise ValueError("Invalid grid indexing used for object")

    def update_E(self, curl_H):
        """custom update equations for inside the object

        Args:
            curl_H: the curl of magnetic field in the grid.

        """
        loc = (self.x, self.y, self.z)
        self.grid.E[loc] += (
            self.grid.courant_number * self.inverse_permittivity * curl_H[loc]
        )

    def update_H(self, curl_E):
        """custom update equations for inside the object

        Args:
            curl_E: the curl of electric field in the grid.

        """

    def __repr__(self):
        return f"{self.__class__.__name__}(name={repr(self.name)})"

    def __str__(self):
        s = "    " + repr(self) + "\n"

        def _handle_slice(s):
            return (
                str(s)
                .replace("slice(", "")
                .replace(")", "")
                .replace(", ", ":")
                .replace("None", "")
            )

        x = _handle_slice(self.x)
        y = _handle_slice(self.y)
        z = _handle_slice(self.z)
        s += f"        @ x={x}, y={y}, z={z}".replace(":,", ",")
        if s[-1] == ":":
            s = s[:-1]
        return s + "\n"


class AbsorbingObject(Object):
    """ An absorbing object takes conductivity into account """

    def __init__(
        self, permittivity: Tensorlike, conductivity: Tensorlike, name: str = None
    ):
        """
        Args:
            permittivity: permittivity tensor
            conductivity: conductivity tensor (will introduce the loss)
            name: name of the object (will become available as attribute to the grid)
        """
        super().__init__(permittivity, name)
        self.conductivity = conductivity

    def _register_grid(
        self, grid: Grid, x: slice = None, y: slice = None, z: slice = None
    ):
        """Register a grid to the object

        Args:
            grid: the grid to register the object into
            x: the x-location of the object in the grid
            y: the y-location of the object in the grid
            z: the z-location of the object in the grid
        """
        super()._register_grid(grid=grid, x=x, y=y, z=z)

        conductivity = bd.asarray(self.conductivity)
        while conductivity.ndim < self.inverse_permittivity.ndim:
            conductivity = conductivity[..., None]
        self.conductivity = bd.broadcast_to(
            conductivity, self.inverse_permittivity.shape
        )

        self.absorption_factor = (
            0.5
            * self.grid.courant_number
            * self.inverse_permittivity
            * self.conductivity
            * self.grid.grid_spacing
            * const.eta0
        )

    def update_E(self, curl_H):
        """custom update equations for inside the absorbing object

        Args:
            curl_H: the curl of magnetic field in the grid.

        """
        loc = (self.x, self.y, self.z)
        self.grid.E[loc] *= (1 - self.absorption_factor) / (1 + self.absorption_factor)
        self.grid.E[loc] += (
            self.grid.courant_number
            * self.inverse_permittivity
            * curl_H[loc]
            / (1 + self.absorption_factor)
        )

    def update_H(self, curl_E):
        """custom update equations for inside the absorbing object

        Args:
            curl_E: the curl of electric field in the grid.

        """


class AnisotropicObject(Object):
    """ An object with anisotropic permittivity tensor """

    def _register_grid(
        self, grid: Grid, x: slice = None, y: slice = None, z: slice = None
    ):
        """Register a grid to the object

        Args:
            grid: the grid to register the object into
            x: the x-location of the object in the grid
            y: the y-location of the object in the grid
            z: the z-location of the object in the grid
        """
        super()._register_grid(grid=grid, x=x, y=y, z=z)
        eye = bd.zeros((self.Nx * self.Ny * self.Nz, 3, 3))
        eye[:, range(3), range(3)] = 1.0
        self.inverse_permittivity = bd.reshape(
            bd.reshape(self.inverse_permittivity, (-1, 1, 3)) * eye,
            (self.Nx, self.Ny, self.Nz, 3, 3),
        )

    def update_E(self, curl_H):
        """custom update equations for inside the anisotropic object

        Args:
            curl_H: the curl of magnetic field in the grid.

        """
        loc = (self.x, self.y, self.z)
        self.grid.E[loc] += bd.reshape(
            self.grid.courant_number
            * bd.bmm(
                bd.reshape(self.inverse_permittivity, (-1, 3, 3)),
                bd.reshape(curl_H[loc], (-1, 3, 1)),
            ),
            (self.Nx, self.Ny, self.Nz, 3),
        )

    def update_H(self, curl_E):
        """custom update equations for inside the anisotropic object

        Args:
            curl_E: the curl of electric field in the grid.

        """



""" Boundaries for the FDTD Grid.

Available Boundaries:

 - PeriodicBoundary
 - PML

"""
## Imports

# typing
from .typing_ import Tensorlike, ListOrSlice, IntOrSlice

# relative
from .grid import Grid
from .backend import backend as bd

## Boundary Conditions [base class]
class Boundary:
    """ an FDTD Boundary [base class] """

    def __init__(self, name: str = None):
        """ Create a boundary

        Args:
            name: name of the boundary
        """
        self.grid = None  # will be registered later
        self.name = name

    def _register_grid(
        self, grid: Grid, x: ListOrSlice, y: ListOrSlice, z: ListOrSlice
    ):
        """ Register a grid to the boundary

        Args:
            grid: the grid to register the boundary to
            x: the x-location of the boundary
            y: the y-location of the boundary
            z: the z-location of the boundary

        Note:
            This is a helper method. To register the boundary to the grid,
            index the grid and assign to it:
            grid[0,:,:] = Boundary(name="boundary_name")
        """
        self.grid = grid
        self.grid.boundaries.append(self)
        self.x = self._handle_slice(x)
        self.y = self._handle_slice(y)
        self.z = self._handle_slice(z)

        #Put into unique_name_check function?
        if self.name is not None:
            if not hasattr(grid, self.name):
                setattr(grid, self.name, self)
            else:
                raise ValueError(
                    f"The grid already has an attribute with name {self.name}"
                )

    def _handle_slice(self, s: ListOrSlice) -> IntOrSlice:
        if isinstance(s, list):
            if len(s) > 1:
                raise ValueError(
                    "Use slices or single numbers to index the grid for a boundary"
                )
            return s[0]
        if isinstance(s, slice):
            if (
                s.start is not None
                and s.stop is not None
                and (s.start == s.stop or abs(s.start - s.stop) == 1)
            ):
                return s.start
            return s
        raise ValueError("Invalid grid indexing used for boundary")

    def update_phi_E(self):
        """ Update convolution [phi_E]

        Note:
            this method is called *before* the electric field is updated
        """

    def update_phi_H(self):
        """ Update convolution [phi_H]

        Note:
            this method is called *before* the magnetic field is updated
        """

    def update_E(self):
        """ Update electric field of the grid

        Note:
            this method is called *after* the grid fields are updated
        """

    def update_H(self):
        """ Update magnetic field of the grid

        Note:
            this method is called *after* the grid fields are updated
        """

    def __repr__(self):
        return f"PML(name={repr(self.name)})"

    def __str__(self):
        s = "    " + repr(self) + "\n"

        def _handle_slice(s):
            return (
                str(s)
                .replace("slice(", "")
                .replace(")", "")
                .replace(", ", ":")
                .replace("None", "")
            )

        x = _handle_slice(self.x)
        y = _handle_slice(self.y)
        z = _handle_slice(self.z)
        s += f"        @ x={x}, y={y}, z={z}".replace(":,", ",")
        if s[-1] == ":":
            s = s[:-1]
        return s + "\n"


## Periodic Boundaries
class PeriodicBoundary(Boundary):
    """ An FDTD Periodic Boundary

    Note:
        Registering a periodic boundary to the grid will change the periodic
        boundary in one of its subclasses: ``_PeriodicBoundaryX``,
        ``_PeriodicBoundaryY`` or ``_PeriodicBoundaryY``, depending on the
        position in the grid.
    """

    def _register_grid(
        self, grid: Grid, x: ListOrSlice, y: ListOrSlice, z: ListOrSlice
    ):
        super()._register_grid(grid=grid, x=x, y=y, z=z)

        if self.x == 0 or self.x == -1:
            self.__class__ = _PeriodicBoundaryX  # subclass of PeriodicBoundary
            if hasattr(grid, "_xlow_boundary") or hasattr(grid, "_xhigh_boundary"):
                raise AttributeError("grid already has an xlow/xhigh boundary!")
            setattr(grid, "_xlow_boundary", self)
            setattr(grid, "_xhigh_boundary", self)
        elif self.y == 0 or self.y == -1:
            self.__class__ = _PeriodicBoundaryY  # subclass of PeriodicBoundary
            if hasattr(grid, "_ylow_boundary") or hasattr(grid, "_yhigh_boundary"):
                raise AttributeError("grid already has an ylow/yhigh boundary!")
            setattr(grid, "_ylow_boundary", self)
            setattr(grid, "_yhigh_boundary", self)
        elif self.z == 0 or self.z == -1:
            self.__class__ = _PeriodicBoundaryZ  # subclass of PeriodicBoundary
            if hasattr(grid, "_zlow_boundary") or hasattr(grid, "_zhigh_boundary"):
                raise AttributeError("grid already has an zlow/zhigh boundary!")
            setattr(grid, "_zlow_boundary", self)
            setattr(grid, "_zhigh_boundary", self)
        else:
            raise IndexError(
                "A periodic boundary should be placed at the boundary of the "
                "grid using a single index (either 0 or -1)"
            )


# Periodic Boundaries in the X-direction
class _PeriodicBoundaryX(PeriodicBoundary):
    def update_E(self):
        """ Update electric field such that periodic boundary conditions in the
        X-direction apply """
        self.grid.E[0, :, :, :] = self.grid.E[-1, :, :, :]

    def update_H(self):
        """ Update magnetic field such that periodic boundary conditions in the
        X-directions apply """
        self.grid.H[-1, :, :, :] = self.grid.H[0, :, :, :]


# Periodic Boundaries in the Y-direction
class _PeriodicBoundaryY(PeriodicBoundary):
    def update_E(self):
        """ Update electric field such that periodic boundary conditions in the
        Y-direction apply """
        self.grid.E[:, 0, :, :] = self.grid.E[:, -1, :, :]

    def update_H(self):
        """ Update magnetic field such that periodic boundary conditions in the
        Y-direction apply """
        self.grid.H[:, -1, :, :] = self.grid.H[:, 0, :, :]


# Periodic Boundaries in the Z-direction
class _PeriodicBoundaryZ(PeriodicBoundary):
    def update_E(self):
        """ Update electric field such that periodic boundary conditions in the
        Z-direction apply """
        self.grid.E[:, :, 0, :] = self.grid.E[:, :, -1, :]

    def update_H(self):
        """ Update magnetic field such that periodic boundary conditions in the
        Z-direction apply """
        self.grid.H[:, :, -1, :] = self.grid.H[:, :, 0, :]


## Perfectly Matched Layer (PML)


class PML(Boundary):
    """ A perfectly matched layer (PML)

    a PML is an impedence-matched area at the boundary of the grid for which
    all fields incident perpendicular to the area are absorbed without
    reflection.

    Note:
        Registering a PML to the grid will monkeypatch the PML to become one of
        its subclasses: ``_PMLXlow``, ``_PMLYlow`` or ``_PMLZlow``,
        ``_PMLXhigh``, ``_PMLYhigh``, ``_PMLZhigh`` depending on the position
        in the grid.
    """

    def __init__(self, a: float = 1e-8, name: str = None):
        """ Perfectly Matched Layer

        Args:
            a: stability parameter
            name: name of the PML
        """
        super().__init__(name=name)

        # TODO: to make this a PML parameter, the *normal* curl equations need to be updated
        self.k = 1.0
        self.thickness = 0

        self.a = a

    def _set_locations(self):
        """ helper function

        sets:
            self.loc: the location of the PML
            self.locx: the location of the PML (only x coordinates)
            self.locy: the location of the PML (only y coordinates)
            self.locz: the location of the PML (only z coordinates)
        """
        raise NotImplementedError

    def _set_shape(self):
        """ helper function

        sets:
            self.shape: the shape of the PML
        """
        raise NotImplementedError

    def _set_sigmaE(self):
        """ helper function

        sets:
            self.sigmaE: the electric conductivity (responsible for the absorption) of
                the PML
        """
        raise NotImplementedError

    def _set_sigmaH(self):
        """ helper function

        sets:
            self.sigmaH: the magnetic conductivity (responsible for the absorption) of
                the PML
        """
        raise NotImplementedError

    def _sigma(self, vect: Tensorlike):
        """ create a cubicly increasing profile for the conductivity """
        return 40 * vect ** 3 / (self.thickness + 1) ** 4

    def _register_grid(
        self, grid: Grid, x: ListOrSlice, y: ListOrSlice, z: ListOrSlice
    ):

        super()._register_grid(grid=grid, x=x, y=y, z=z)

        if (
            (self.x.start is None or self.x.start == 0)
            and (self.x.stop is not None)
            and (self.x.stop > 0)
        ):
            self.__class__ = _PMLXlow
            if hasattr(grid, "_xlow_boundary"):
                raise AttributeError("grid already has an xlow boundary!")
            setattr(grid, "_xlow_boundary", self)
            self._calculate_parameters(thickness=self.x.stop)
        elif (
            (self.x.start is not None) and (self.x.stop is None) and (self.x.start < 0)
        ):
            self.__class__ = _PMLXhigh
            if hasattr(grid, "_xhigh_boundary"):
                raise AttributeError("grid already has an xhigh boundary!")
            setattr(grid, "_xhigh_boundary", self)
            self._calculate_parameters(thickness=-self.x.start)
        elif (
            (self.y.start is None or self.y.start == 0)
            and (self.y.stop is not None)
            and (self.y.stop > 0)
        ):
            self.__class__ = _PMLYlow
            if hasattr(grid, "_ylow_boundary"):
                raise AttributeError("grid already has an ylow boundary!")
            setattr(grid, "_ylow_boundary", self)
            self._calculate_parameters(thickness=self.y.stop)
        elif (
            (self.y.start is not None) and (self.y.stop is None) and (self.y.start < 0)
        ):
            self.__class__ = _PMLYhigh
            if hasattr(grid, "_yhigh_boundary"):
                raise AttributeError("grid already has an yhigh boundary!")
            setattr(grid, "_yhigh_boundary", self)
            self._calculate_parameters(thickness=-self.y.start)
        elif (
            (self.z.start is None or self.z.start == 0)
            and (self.z.stop is not None)
            and (self.z.stop > 0)
        ):
            self.__class__ = _PMLZlow
            if hasattr(grid, "_zlow_boundary"):
                raise AttributeError("grid already has an zlow boundary!")
            setattr(grid, "_zlow_boundary", self)
            self._calculate_parameters(thickness=self.z.stop)
        elif (
            (self.z.start is not None) and (self.z.stop is None) and (self.z.start < 0)
        ):
            self.__class__ = _PMLZhigh
            if hasattr(grid, "_zhigh_boundary"):
                raise AttributeError("grid already has an zhigh boundary!")
            setattr(grid, "_zhigh_boundary", self)
            self._calculate_parameters(thickness=-self.z.start)
        else:
            raise IndexError(
                "not a valid slice for a PML. Make sure the slice is at the border of the PML"
            )

    def _handle_slice(self, s: ListOrSlice) -> slice:
        if isinstance(s, list):
            raise ValueError("One can only use slices to index the grid for a PML")
        if isinstance(s, slice):
            return s
        raise ValueError("Invalid grid indexing used for boundary")

    def _calculate_parameters(self, thickness: int = 10):
        """ Calculate the parameters for the PML

        Args:
            thickness=10: The thickness of the PML. The thickness can be specified as
                integer [gridpoints] or as float [meters].
        """

        self.thickness = thickness

        # set orientation dependent parameters: (different for x, y, z-PML)
        # NOTE: these methods are implemented by the subclasses of PML.
        self._set_locations()
        self._set_shape()
        self._set_sigmaE()
        self._set_sigmaH()

        # set the other parameters
        Nx, Ny, Nz = self.shape  # is defined by _set_shape()
        self.phi_E = bd.zeros((Nx, Ny, Nz, 3))
        self.phi_H = bd.zeros((Nx, Ny, Nz, 3))
        self.psi_Ex = bd.zeros((Nx, Ny, Nz, 3))
        self.psi_Ey = bd.zeros((Nx, Ny, Nz, 3))
        self.psi_Ez = bd.zeros((Nx, Ny, Nz, 3))
        self.psi_Hx = bd.zeros((Nx, Ny, Nz, 3))
        self.psi_Hy = bd.zeros((Nx, Ny, Nz, 3))
        self.psi_Hz = bd.zeros((Nx, Ny, Nz, 3))

        self.bE = bd.exp(-(self.sigmaE / self.k + self.a) * self.grid.courant_number)
        self.cE = (
            (self.bE - 1.0)
            * self.sigmaE  # is defined by _set_sigmaE()
            / (self.sigmaE * self.k + self.a * self.k ** 2)
        )

        self.bH = bd.exp(-(self.sigmaH / self.k + self.a) * self.grid.courant_number)
        self.cH = (
            (self.bH - 1.0)
            * self.sigmaH  # is defined by _set_sigmaH()
            / (self.sigmaH * self.k + self.a * self.k ** 2)
        )

    def update_E(self):
        """ Update electric field of the grid

        Note:
            this method is called *after* the electric field is updated
        """
        self.grid.E[self.loc] += (
            self.grid.courant_number
            * self.grid.inverse_permittivity[self.loc]
            * self.phi_E
        )

    def update_H(self):
        """ Update magnetic field of the grid

        Note:
            this method is called *after* the magnetic field is updated
        """
        self.grid.H[self.loc] -= (
            self.grid.courant_number
            * self.grid.inverse_permeability[self.loc]
            * self.phi_H
        )

    def update_phi_E(self):
        """ Update convolution [phi_E]

        Note:
            this method is called *before* the electric field is updated
        """
        self.psi_Ex *= self.bE
        self.psi_Ey *= self.bE
        self.psi_Ez *= self.bE

        c = self.cE
        Hx = self.grid.H[self.locx]
        Hy = self.grid.H[self.locy]
        Hz = self.grid.H[self.locz]

        self.psi_Ex[:, 1:, :, 1] += (Hz[:, 1:, :] - Hz[:, :-1, :]) * c[:, 1:, :, 1]
        self.psi_Ex[:, :, 1:, 2] += (Hy[:, :, 1:] - Hy[:, :, :-1]) * c[:, :, 1:, 2]

        self.psi_Ey[:, :, 1:, 2] += (Hx[:, :, 1:] - Hx[:, :, :-1]) * c[:, :, 1:, 2]
        self.psi_Ey[1:, :, :, 0] += (Hz[1:, :, :] - Hz[:-1, :, :]) * c[1:, :, :, 0]

        self.psi_Ez[1:, :, :, 0] += (Hy[1:, :, :] - Hy[:-1, :, :]) * c[1:, :, :, 0]
        self.psi_Ez[:, 1:, :, 1] += (Hx[:, 1:, :] - Hx[:, :-1, :]) * c[:, 1:, :, 1]

        self.phi_E[..., 0] = self.psi_Ex[..., 1] - self.psi_Ex[..., 2]
        self.phi_E[..., 1] = self.psi_Ey[..., 2] - self.psi_Ey[..., 0]
        self.phi_E[..., 2] = self.psi_Ez[..., 0] - self.psi_Ez[..., 1]

    def update_phi_H(self):
        """ Update convolution [phi_H]

        Note:
            this method is called *before* the magnetic field is updated
        """
        self.psi_Hx *= self.bH
        self.psi_Hy *= self.bH
        self.psi_Hz *= self.bH

        c = self.cH
        Ex = self.grid.E[self.locx]
        Ey = self.grid.E[self.locy]
        Ez = self.grid.E[self.locz]

        self.psi_Hx[:, :-1, :, 1] += (Ez[:, 1:, :] - Ez[:, :-1, :]) * c[:, :-1, :, 1]
        self.psi_Hx[:, :, :-1, 2] += (Ey[:, :, 1:] - Ey[:, :, :-1]) * c[:, :, :-1, 2]

        self.psi_Hy[:, :, :-1, 2] += (Ex[:, :, 1:] - Ex[:, :, :-1]) * c[:, :, :-1, 2]
        self.psi_Hy[:-1, :, :, 0] += (Ez[1:, :, :] - Ez[:-1, :, :]) * c[:-1, :, :, 0]

        self.psi_Hz[:-1, :, :, 0] += (Ey[1:, :, :] - Ey[:-1, :, :]) * c[:-1, :, :, 0]
        self.psi_Hz[:, :-1, :, 1] += (Ex[:, 1:, :] - Ex[:, :-1, :]) * c[:, :-1, :, 1]

        self.phi_H[..., 0] = self.psi_Hx[..., 1] - self.psi_Hx[..., 2]
        self.phi_H[..., 1] = self.psi_Hy[..., 2] - self.psi_Hy[..., 0]
        self.phi_H[..., 2] = self.psi_Hz[..., 0] - self.psi_Hz[..., 1]


class _PMLXlow(PML):
    """ A perfectly matched layer to place where X is low. """

    def _set_locations(self):
        self.loc = (slice(None, self.thickness), slice(None), slice(None), slice(None))
        self.locx = (slice(None, self.thickness), slice(None), slice(None), 0)
        self.locy = (slice(None, self.thickness), slice(None), slice(None), 1)
        self.locz = (slice(None, self.thickness), slice(None), slice(None), 2)

    def _set_shape(self):
        self.shape = (self.thickness, self.grid.Ny, self.grid.Nz)

    def _set_sigmaE(self):
        sigma = self._sigma(bd.arange(self.thickness - 0.5, -0.5, -1.0))
        self.sigmaE = bd.zeros((self.thickness, self.grid.Ny, self.grid.Nz, 3))
        self.sigmaE[:, :, :, 0] = sigma[:, None, None]

    def _set_sigmaH(self):
        sigma = self._sigma(bd.arange(self.thickness - 1.0, 0, -1.0))
        self.sigmaH = bd.zeros((self.thickness, self.grid.Ny, self.grid.Nz, 3))
        self.sigmaH[:-1, :, :, 0] = sigma[:, None, None]


class _PMLXhigh(PML):
    """ A perfectly matched layer to place where X is high. """

    def _set_locations(self):
        self.loc = (slice(-self.thickness, None), slice(None), slice(None), slice(None))
        self.locx = (slice(-self.thickness, None), slice(None), slice(None), 0)
        self.locy = (slice(-self.thickness, None), slice(None), slice(None), 1)
        self.locz = (slice(-self.thickness, None), slice(None), slice(None), 2)

    def _set_shape(self):
        self.shape = (self.thickness, self.grid.Ny, self.grid.Nz)

    def _set_sigmaE(self):
        sigma = self._sigma(bd.arange(0.5, self.thickness + 0.5, 1.0))
        self.sigmaE = bd.zeros((self.thickness, self.grid.Ny, self.grid.Nz, 3))
        self.sigmaE[:, :, :, 0] = sigma[:, None, None]

    def _set_sigmaH(self):
        sigma = self._sigma(bd.arange(1.0, self.thickness, 1.0))
        self.sigmaH = bd.zeros((self.thickness, self.grid.Ny, self.grid.Nz, 3))
        self.sigmaH[:-1, :, :, 0] = sigma[:, None, None]


class _PMLYlow(PML):
    """ A perfectly matched layer to place where Y is low. """

    def _set_locations(self):
        self.loc = (slice(None), slice(None, self.thickness), slice(None))
        self.locx = (slice(None), slice(None, self.thickness), slice(None), 0)
        self.locy = (slice(None), slice(None, self.thickness), slice(None), 1)
        self.locz = (slice(None), slice(None, self.thickness), slice(None), 2)

    def _set_shape(self):
        self.shape = (self.grid.Nx, self.thickness, self.grid.Nz)

    def _set_sigmaE(self):
        sigma = self._sigma(bd.arange(self.thickness - 0.5, -0.5, -1.0))
        self.sigmaE = bd.zeros((self.grid.Nx, self.thickness, self.grid.Nz, 3))
        self.sigmaE[:, :, :, 1] = sigma[None, :, None]

    def _set_sigmaH(self):
        sigma = self._sigma(bd.arange(self.thickness - 1.0, 0, -1.0))
        self.sigmaH = bd.zeros((self.grid.Nx, self.thickness, self.grid.Nz, 3))
        self.sigmaH[:, :-1, :, 1] = sigma[None, :, None]


class _PMLYhigh(PML):
    """ A perfectly matched layer to place where Y is high. """

    def _set_locations(self):
        self.loc = (slice(None), slice(-self.thickness, None), slice(None), slice(None))
        self.locx = (slice(None), slice(-self.thickness, None), slice(None), 0)
        self.locy = (slice(None), slice(-self.thickness, None), slice(None), 1)
        self.locz = (slice(None), slice(-self.thickness, None), slice(None), 2)

    def _set_shape(self):
        self.shape = (self.grid.Nx, self.thickness, self.grid.Nz)

    def _set_sigmaE(self):
        sigma = self._sigma(bd.arange(0.5, self.thickness + 0.5, 1.0))
        self.sigmaE = bd.zeros((self.grid.Nx, self.thickness, self.grid.Nz, 3))
        self.sigmaE[:, :, :, 1] = sigma[None, :, None]

    def _set_sigmaH(self):
        sigma = self._sigma(bd.arange(1.0, self.thickness, 1.0))
        self.sigmaH = bd.zeros((self.grid.Nx, self.thickness, self.grid.Nz, 3))
        self.sigmaH[:, :-1, :, 1] = sigma[None, :, None]


class _PMLZlow(PML):
    """ A perfectly matched layer to place where Z is low. """

    def _set_locations(self):
        self.loc = (slice(None), slice(None), slice(None, self.thickness), slice(None))
        self.locx = (slice(None), slice(None), slice(None, self.thickness), 0)
        self.locy = (slice(None), slice(None), slice(None, self.thickness), 1)
        self.locz = (slice(None), slice(None), slice(None, self.thickness), 2)

    def _set_shape(self):
        self.shape = (self.grid.Nx, self.grid.Ny, self.thickness)

    def _set_sigmaE(self):
        sigma = self._sigma(bd.arange(self.thickness - 0.5, -0.5, -1.0))
        self.sigmaE = bd.zeros((self.grid.Nx, self.grid.Ny, self.thickness, 3))
        self.sigmaE[:, :, :, 2] = sigma[None, None, :]

    def _set_sigmaH(self):
        sigma = self._sigma(bd.arange(self.thickness - 1.0, 0, -1.0))
        self.sigmaH = bd.zeros((self.grid.Nx, self.grid.Ny, self.thickness, 3))
        self.sigmaH[:, :, :-1, 2] = sigma[None, None, :]


class _PMLZhigh(PML):
    """ A perfectly matched layer to place where Z is high. """

    def _set_locations(self):
        self.loc = (slice(None), slice(None), slice(-self.thickness, None), slice(None))
        self.locx = (slice(None), slice(None), slice(-self.thickness, None), 0)
        self.locy = (slice(None), slice(None), slice(-self.thickness, None), 1)
        self.locz = (slice(None), slice(None), slice(-self.thickness, None), 2)

    def _set_shape(self):
        self.shape = (self.grid.Nx, self.grid.Ny, self.thickness)

    def _set_sigmaE(self):
        sigma = self._sigma(bd.arange(0.5, self.thickness + 0.5, 1.0))
        self.sigmaE = bd.zeros((self.grid.Nx, self.grid.Ny, self.thickness, 3))
        self.sigmaE[:, :, :, 2] = sigma[None, None, :]

    def _set_sigmaH(self):
        sigma = self._sigma(bd.arange(1.0, self.thickness, 1.0))
        self.sigmaH = bd.zeros((self.grid.Nx, self.grid.Ny, self.thickness, 3))
        self.sigmaH[:, :, :-1, 2] = sigma[None, None, :]


def DomainBorderPML(grid, border_cells=5):
    '''
    Some problem setups require a layer of PML all the way around the problem.
    This is a convenience function to add such a layer to an existing grid.
    Caution: Alters grid in-place.
    '''
    # Doesn't return anything, unlike other .boundaries functions - is that breaking the API?
    # There might be a more straightforward numpy slicing solution...
    # maybe with
    # raw = np.pad(raw, [(pcb.pml_cells, pcb.pml_cells), (pcb.pml_cells, pcb.pml_cells),
    #    (pcb.pml_cells, pcb.pml_cells)], mode='constant') and then indexing the grid by that?

    if(grid.Nx < border_cells*2 or grid.Ny < border_cells*2 or grid.Nz < border_cells*2):
        raise IndexError("PML border_cells larger than domain!")

    #overly-complicated slicing doesn't overlap,
    #haven't checked if that's important

    #top and bottom
    grid[:, : ,0:border_cells] = PML()
    grid[:, : ,-border_cells:] = PML()

    #left and right
    grid[0:border_cells, :, border_cells:-border_cells] = PML()
    grid[-border_cells:, :, border_cells:-border_cells] = PML()

    #front and back
    grid[border_cells:-border_cells, 0:border_cells, border_cells:-border_cells] = PML()
    grid[border_cells:-border_cells, -border_cells:, border_cells:-border_cells] = PML()



""" Detectors for the FDTD Grid.

Available Detectors:

 - LineDetector

"""

## Imports

# typing
from .typing_ import ListOrSlice, Tuple, List

# relative
from .grid import Grid
from .backend import backend as bd
from .constants import X, Y, Z

## Detector
class LineDetector:
    """ A detector along a line in the FDTD grid """

    def __init__(self, name=None):
        """Create a line detector

        Args:
            name: name of the Detector

        """
        self.grid = None
        self.E = []
        self.H = []
        self.name = name

    def _register_grid(
        self, grid: Grid, x: ListOrSlice, y: ListOrSlice, z: ListOrSlice
    ):
        """Register a grid to the detector

        Args:
            grid: the grid to place the detector into
            x: the x-location of the detector in the grid
            y: the y-location of the detector in the grid
            z: the z-location of the detector in the grid

        Note:
            As its name suggests, this detector is a LINE detector.
            Hence the detector spans the diagonal of the cube
            defined by the slices in the grid.
        """
        self.grid = grid
        self.grid.detectors.append(self)
        if self.name is not None:
            if not hasattr(grid, self.name):
                setattr(grid, self.name, self)
            else:
                raise ValueError(
                    f"The grid already has an attribute with name {self.name}"
                )

        self.x, self.y, self.z = self._handle_slices(x, y, z)

    def _handle_slices(
        self, x: ListOrSlice, y: ListOrSlice, z: ListOrSlice
    ) -> Tuple[List, List, List]:
        """Convert slices in the grid to lists

        This is necessary to make the source span the diagonal of the volume
        defined by the slices.

        Args:
            x: The x-location of the volume in the grid
            y: The y-location of the volume in the grid
            z: The z-location of the volume in the grid

        Returns:
            x, y, z: the x, y and z coordinates of the source as lists

        """

        # if list-indices were chosen:
        if isinstance(x, list) and isinstance(y, list) and isinstance(z, list):
            if len(x) != len(y) or len(y) != len(z) or len(z) != len(x):
                raise IndexError(
                    "sources require grid to be indexed with slices or equal length list-indices"
                )
            return x, y, z

        # if a combination of list-indices and slices were chosen,
        # convert the list-indices to slices.
        # TODO: maybe issue a warning here?
        if isinstance(x, list):
            x = slice(x[0], x[-1], None)
        if isinstance(y, list):
            y = slice(y[0], y[-1], None)
        if isinstance(z, list):
            z = slice(z[0], z[-1], None)

        # if we get here, we can assume slices:
        x0 = x.start if x.start is not None else 0
        y0 = y.start if y.start is not None else 0
        z0 = z.start if z.start is not None else 0
        x1 = x.stop if x.stop is not None else self.grid.Nx
        y1 = y.stop if y.stop is not None else self.grid.Ny
        z1 = z.stop if z.stop is not None else self.grid.Nz

        # we can now convert these coordinates into index lists
        m = max(abs(x1 - x0), abs(y1 - y0), abs(z1 - z0))
        x = [v.item() for v in bd.array(bd.linspace(x0, x1, m, endpoint=False), bd.int)]
        y = [v.item() for v in bd.array(bd.linspace(y0, y1, m, endpoint=False), bd.int)]
        z = [v.item() for v in bd.array(bd.linspace(z0, z1, m, endpoint=False), bd.int)]
        return x, y, z

    def detect_E(self):
        """ detect the electric field at a certain location in the grid """
        # TODO: there is a performance bottleneck here (indexing with lists)
        E = self.grid.E[self.x, self.y, self.z]
        self.E.append(E)

    def detect_H(self):
        """ detect the magnetic field at a certain location in the grid """
        # TODO: there is a performance bottleneck here (indexing with lists)
        H = self.grid.H[self.x, self.y, self.z]
        self.H.append(H)

    def __repr__(self):
        return f"{self.__class__.__name__}(name={repr(self.name)})"

    def __str__(self):
        s = "    " + repr(self) + "\n"
        x = f"[{self.x[0]}, ... , {self.x[-1]}]"
        y = f"[{self.y[0]}, ... , {self.y[-1]}]"
        z = f"[{self.z[0]}, ... , {self.z[-1]}]"
        s += f"        @ x={x}, y={y}, z={z}\n"
        return s

    def detector_values(self):
        """ outputs what detector detects """
        return {"E": self.E, "H": self.H}


# is the "detector" paradigm necessary? Can we just flag a segment of the base mesh to be
# stored per timestep?

## BlockDetector
class BlockDetector:
    """ A detector along a block in the FDTD grid """

    """ Basic copy of LineDetector code, changed detect functions """

    def __init__(self, name=None):
        """Create a block detector

        Args:
            name: name of the Detector

        """
        self.grid = None
        self.E = []
        self.H = []
        self.name = name

    def _register_grid(
        self, grid: Grid, x: ListOrSlice, y: ListOrSlice, z: ListOrSlice
    ):
        """Register a grid to the detector

        Args:
            grid: the grid to place the detector into
            x: the x-location of the detector in the grid
            y: the y-location of the detector in the grid
            z: the z-location of the detector in the grid

        Note:
            As its name suggests, this detector is a BLOCK detector.
            Hence the detector spans the VOLUME of the cube
            defined by the slices in the grid.
        """
        self.grid = grid
        self.grid.detectors.append(self)
        if self.name is not None:
            if not hasattr(grid, self.name):
                setattr(grid, self.name, self)
            else:
                raise ValueError(
                    f"The grid already has an attribute with name {self.name}"
                )

        self.x, self.y, self.z = self._handle_slices(x, y, z)

    def _handle_slices(
        self, x: ListOrSlice, y: ListOrSlice, z: ListOrSlice
    ) -> Tuple[List, List, List]:
        """Convert slices in the grid to lists

        This is necessary to make the source span the volume
        defined by the slices.

        Args:
            x: The x-location of the volume in the grid
            y: The y-location of the volume in the grid
            z: The z-location of the volume in the grid

        Returns:
            x, y, z: the x, y and z coordinates of the source as lists

        """

        # if list-indices were chosen:
        if isinstance(x, list) and isinstance(y, list) and isinstance(z, list):
            if len(x) != len(y) or len(y) != len(z) or len(z) != len(x):
                raise IndexError(
                    "sources require grid to be indexed with slices or equal length list-indices"
                )
            return x, y, z

        # if a combination of list-indices and slices were chosen,
        # convert the list-indices to slices.
        # TODO: maybe issue a warning here?
        if isinstance(x, list):
            x = slice(x[0], x[-1], None)
        if isinstance(y, list):
            y = slice(y[0], y[-1], None)
        if isinstance(z, list):
            z = slice(z[0], z[-1], None)

        # if we get here, we can assume slices:
        x0 = x.start if x.start is not None else 0
        y0 = y.start if y.start is not None else 0
        z0 = z.start if z.start is not None else 0
        x1 = x.stop if x.stop is not None else self.grid.Nx
        y1 = y.stop if y.stop is not None else self.grid.Ny
        z1 = z.stop if z.stop is not None else self.grid.Nz

        # we can now convert these coordinates into index lists
        x = [v.item() for v in bd.arange(x0, x1 + 1)]
        y = [v.item() for v in bd.arange(y0, y1 + 1)]
        z = [v.item() for v in bd.arange(z0, z1 + 1)]
        return x, y, z

    def detect_E(self):
        """ detect the electric field at a certain location in the grid """
        # TODO: there is a performance bottleneck here (indexing with lists)
        E = []
        for i, row in enumerate(self.x):
            E.append([])
            for j, col in enumerate(self.y):
                E[i].append([])
                for pillar in self.z:
                    E[i][j].append(self.grid.E[row, col, [pillar]][0])
        # E = self.grid.E[self.x, self.y, self.z]
        self.E.append(E)

    def detect_H(self):
        """ detect the magnetic field at a certain location in the grid """
        # TODO: there is a performance bottleneck here (indexing with lists)
        H = []
        for i, row in enumerate(self.x):
            H.append([])
            for j, col in enumerate(self.y):
                H[i].append([])
                for pillar in self.z:
                    H[i][j].append(self.grid.H[row, col, [pillar]][0])
        # H = self.grid.H[self.x, self.y, self.z]
        self.H.append(H)

    def __repr__(self):
        return f"{self.__class__.__name__}(name={repr(self.name)})"

    def __str__(self):
        s = "    " + repr(self) + "\n"
        x = f"[{self.x[0]}, ... , {self.x[-1]}]"
        y = f"[{self.y[0]}, ... , {self.y[-1]}]"
        z = f"[{self.z[0]}, ... , {self.z[-1]}]"
        s += f"        @ x={x}, y={y}, z={z}\n"
        return s

    def detector_values(self):
        """ outputs what detector detects """
        return {"E": self.E, "H": self.H}


## CurrentDetector
class CurrentDetector:
    """ A current detector. """

    """



    """

    # Alternative APIs:
    #
    # 1. Current detection could be a standard detector output,
    # if every h-field probe pointed two ways.
    # 2. No current detector class at all. SoftArbitraryPointSource
    # makes two H-detectors. (can't test separated then).
    # Don't know if this op over a volume is valid. Should be.
    # TODO: Detector geometry as an argument?
    # Lots of interesting I detector geometries that might be worth implementing,
    # ("integrated current loop"), this could be a flag here.
    #
    # Existing sources are all z-polarized; E-detectors output all the polarizations.
    # Unlike other Detectors, this does not yet output a polarized current!
    #
    # *Source* polarization is probably an important feature for 3D electrical work
    # not sure how important it is for optical stuff

    def __init__(self, name=None):
        """Create a block detector

        Args:
            name: name of the Detector

        """
        self.grid = None
        self.orientation = None
        self.I = []
        self.name = name

    def _register_grid(
        self, grid: Grid, x: ListOrSlice, y: ListOrSlice, z: ListOrSlice
    ):
        """Register a grid to the detector

        Args:
            grid: the grid to place the detector into
            x: the x-location of the detector in the grid
            y: the y-location of the detector in the grid
            z: the z-location of the detector in the grid
            orientation: a unit vector (a length 1) describing the orientation?

        Note:
            As its name suggests, this detector is a BLOCK detector.
            Hence the detector spans the VOLUME of the cube
            defined by the slices in the grid.
        """
        self.grid = grid
        self.grid.detectors.append(self)
        if self.name is not None:
            if not hasattr(grid, self.name):
                setattr(grid, self.name, self)
            else:
                raise ValueError(
                    f"The grid already has an attribute with name {self.name}"
                )

        self.x, self.y, self.z = self._handle_slices(x, y, z)

    def _handle_slices(
        self, x: ListOrSlice, y: ListOrSlice, z: ListOrSlice
    ) -> Tuple[List, List, List]:
        """Convert slices in the grid to lists

        This is necessary to make the source span the volume
        defined by the slices.

        Args:
            x: The x-location of the volume in the grid
            y: The y-location of the volume in the grid
            z: The z-location of the volume in the grid

        Returns:
            x, y, z: the x, y and z coordinates of the source as lists

        """

        # if list-indices were chosen:
        if isinstance(x, list) and isinstance(y, list) and isinstance(z, list):
            if len(x) != len(y) or len(y) != len(z) or len(z) != len(x):
                raise IndexError(
                    "sources require grid to be indexed with slices or equal length list-indices"
                )
            return x, y, z

        # if a combination of list-indices and slices were chosen,
        # convert the list-indices to slices.
        # TODO: maybe issue a warning here?
        if isinstance(x, list):
            x = slice(x[0], x[-1], None)
        if isinstance(y, list):
            y = slice(y[0], y[-1], None)
        if isinstance(z, list):
            z = slice(z[0], z[-1], None)

        # if we get here, we can assume slices:
        x0 = x.start if x.start is not None else 0
        y0 = y.start if y.start is not None else 0
        z0 = z.start if z.start is not None else 0
        x1 = x.stop if x.stop is not None else self.grid.Nx
        y1 = y.stop if y.stop is not None else self.grid.Ny
        z1 = z.stop if z.stop is not None else self.grid.Nz

        # we can now convert these coordinates into index lists
        x = [v.item() for v in bd.arange(x0, x1 + 1)]
        y = [v.item() for v in bd.arange(y0, y1 + 1)]
        z = [v.item() for v in bd.arange(z0, z1 + 1)]
        return x, y, z

    def detect_E(self):
        """ detect the electric field at a certain location in the grid """

    def single_point_current(self, px, py, pz):
        """

        Only Z-polarized for now. Can probably do a cross product to get arbitrary polarizations

        ^
        |
        |
        X---->

        TODO: FIXME: IMPORTANT:
        material magnetic permeability? find test cases!



        Implements the first correction from [Fang 1994] (two
        cells are spatially averaged to account for Yee cell half-step inaccuracies),
        but not the second one (minor loss of accuracy).

        Jiayuan Fang, Danwei Xue.
        Precautions in the calculation of impedance in FDTD computations.
        Proceedings of IEEE Antennas and Propagation Society International Symposium
        and URSI National Radio Science Meeting, vol. 3, 1994, p. 1814–7 vol.3.
        https://doi.org/10.1109/APS.1994.408185.


        Luebbers RJ, Langdon HS.
        A simple feed model that reduces time steps needed for
        FDTD antenna and microstrip calculations.
        IEEE Trans Antennas Propagat 1996;44:1000–5.
        https://doi.org/10.1109/8.504308.

        """

        # [Luebbers 1996 1992]
        # /sqrt(mu_0)s removed!

        # option for unit factor here? Units will get very complicated otherwise

        current_vector_1 = (
            (self.grid.H[px, py - 1, pz, X]) - (self.grid.H[px, py, pz, X])
        ) * self.grid.grid_spacing
        current_vector_2 = (
            (self.grid.H[px, py, pz, Y]) - (self.grid.H[px - 1, py, pz, Y])
        ) * self.grid.grid_spacing

        current_1 = current_vector_1 + current_vector_2
        # current_1 = float(current.cpu())

        current_vector_1 = (
            (self.grid.H[px, py - 1, pz - 1, X]) - (self.grid.H[px, py, pz - 1, X])
        ) * self.grid.grid_spacing
        current_vector_2 += (
            (self.grid.H[px, py, pz - 1, Y]) - (self.grid.H[px - 1, py, pz - 1, Y])
        ) * self.grid.grid_spacing
        # current_2 = float(current_2.cpu())
        current_2 = current_vector_1 + current_vector_2

        I = (current_1 + current_2) / 2.0

        return I

    def detect_H(self):

        # should detector outputs be bd.array() by default?
        # for that matter, should they always be converted to numpy?

        I = []
        for i, row in enumerate(self.x):
            I.append([])
            for j, col in enumerate(self.y):
                I[i].append([])
                for k, pillar in enumerate(self.z):
                    I[i][j].append([])
                    I[i][j][k] = self.single_point_current(row, col, pillar)

        self.I.append(I)

    # can these functions be templated somehow?
    def __repr__(self):
        return f"{self.__class__.__name__}(name={repr(self.name)})"

    def __str__(self):
        s = "    " + repr(self) + "\n"
        x = f"[{self.x[0]}, ... , {self.x[-1]}]"
        y = f"[{self.y[0]}, ... , {self.y[-1]}]"
        z = f"[{self.z[0]}, ... , {self.z[-1]}]"
        s += f"        @ x={x}, y={y}, z={z}\n"
        return s

    def detector_values(self):
        """ outputs what detector detects """
        return {"I": self.I}



""" Python 3D FDTD Simulator """

__author__ = "Floris laporte"
__version__ = "0.2.7"

from .grid import Grid
from .sources import PointSource, LineSource, PlaneSource
from .detectors import LineDetector, BlockDetector, CurrentDetector
from .objects import Object, AbsorbingObject, AnisotropicObject
from .boundaries import PeriodicBoundary, PML
from .backend import backend
from .backend import set_backend
from .fourier import FrequencyRoutines
from .visualization import dB_map_2D, plot_detection



""" Selects the backend for the fdtd-package.

The `fdtd` library allows to choose a backend. The ``numpy`` backend is the
default one, but there are also several additional PyTorch backends:

    - ``numpy`` (defaults to float64 arrays)
    - ``torch`` (defaults to float64 tensors)
    - ``torch.float32``
    - ``torch.float64``
    - ``torch.cuda`` (defaults to float64 tensors)
    - ``torch.cuda.float32``
    - ``torch.cuda.float64``

For example, this is how to choose the `"torch"` backend: ::

    fdtd.set_backend("torch")

In general, the ``numpy`` backend is preferred for standard CPU calculations
with `"float64"` precision. In general, ``float64`` precision is always
preferred over ``float32`` for FDTD simulations, however, ``float32`` might
give a significant performance boost.

The ``cuda`` backends are only available for computers with a GPU.

"""

## Imports

# Numpy Backend
import numpy  # numpy has to be present
from functools import wraps

# used only by test runner.
# default must be idx 0.
backend_names = [
    dict(backends="numpy"),
    dict(backends="torch.float32"),
    dict(backends="torch.float64"),
    dict(backends="torch.cuda.float32"),
    dict(backends="torch.cuda.float64"),
]

numpy_float_dtypes = {
    getattr(numpy, "float_", numpy.float64),
    getattr(numpy, "float16", numpy.float64),
    getattr(numpy, "float32", numpy.float64),
    getattr(numpy, "float64", numpy.float64),
    getattr(numpy, "float128", numpy.float64),
}


# Torch Backends (and flags)
try:
    import torch

    torch.set_default_dtype(torch.float64)  # we need more precision for FDTD
    try:  # we don't need gradients (for now)
        torch._C.set_grad_enabled(False)  # type: ignore
    except AttributeError:
        torch._C._set_grad_enabled(False)
    TORCH_AVAILABLE = True
    TORCH_CUDA_AVAILABLE = torch.cuda.is_available()
except ImportError:
    TORCH_AVAILABLE = False
    TORCH_CUDA_AVAILABLE = False


# Base Class
class Backend:
    """Backend Base Class"""

    # constants
    pi = numpy.pi

    def __repr__(self):
        return self.__class__.__name__


def _replace_float(func):
    """replace the default dtype a function is called with"""

    @wraps(func)
    def new_func(self, *args, **kwargs):
        result = func(*args, **kwargs)
        if result.dtype in numpy_float_dtypes:
            result = numpy.asarray(result, dtype=self.float)
        return result

    return new_func


# Numpy Backend
class NumpyBackend(Backend):
    """Numpy Backend"""

    # types
    int = numpy.int64
    """ integer type for array"""

    float = numpy.float64
    """ floating type for array """

    # methods
    asarray = _replace_float(numpy.asarray)

    exp = staticmethod(numpy.exp)
    """ exponential of all elements in array """

    sin = staticmethod(numpy.sin)
    """ sine of all elements in array """

    cos = staticmethod(numpy.cos)
    """ cosine of all elements in array """

    sum = staticmethod(numpy.sum)
    """ sum elements in array """

    max = staticmethod(numpy.max)
    """ max element in array """

    stack = staticmethod(numpy.stack)
    """ stack multiple arrays """

    transpose = staticmethod(numpy.transpose)
    """ transpose array by flipping two dimensions """

    reshape = staticmethod(numpy.reshape)
    """ reshape array into given shape """

    squeeze = staticmethod(numpy.squeeze)
    """ remove dim-1 dimensions """

    broadcast_arrays = staticmethod(numpy.broadcast_arrays)
    """ broadcast arrays """

    broadcast_to = staticmethod(numpy.broadcast_to)
    """ broadcast array into shape """

    @staticmethod
    def bmm(arr1, arr2):
        """batch matrix multiply two arrays"""
        return numpy.einsum("ijk,ikl->ijl", arr1, arr2)

    @staticmethod
    def is_array(arr):
        """check if an object is an array"""
        return isinstance(arr, numpy.ndarray)

    # constructors
    array = _replace_float(numpy.array)
    """ create an array from an array-like sequence """

    ones = _replace_float(numpy.ones)
    """ create an array filled with ones """

    zeros = _replace_float(numpy.zeros)
    """ create an array filled with zeros """

    zeros_like = staticmethod(numpy.zeros_like)
    """ create an array filled with zeros """

    linspace = _replace_float(numpy.linspace)
    """ create a linearly spaced array between two points """

    arange = _replace_float(numpy.arange)
    """ create a range of values """

    pad = staticmethod(numpy.pad)

    fftfreq = staticmethod(numpy.fft.fftfreq)

    fft = staticmethod(numpy.fft.fft)

    exp = staticmethod(numpy.exp)

    divide = staticmethod(numpy.divide)

    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # beware to future people:
    # because this line *redefines numpy*,
    # you have to add your new staticmethods /above/ this line to avoid mystification.
    # <3 <3 <3 <3
    #
    # could this (and below) perhaps be changed to "to_numpy()"
    # or maybe "check_numpy" ?
    numpy = _replace_float(numpy.asarray)
    """ convert the array to numpy array """


# Torch Backend
if TORCH_AVAILABLE:
    import torch

    class TorchBackend(Backend):
        """Torch Backend"""

        # types
        int = torch.int64
        """ integer type for array"""

        float = torch.get_default_dtype()
        """ floating type for array """

        # methods
        asarray = staticmethod(torch.as_tensor)
        """ create an array """

        exp = staticmethod(torch.exp)
        """ exponential of all elements in array """

        sin = staticmethod(torch.sin)
        """ sine of all elements in array """

        cos = staticmethod(torch.cos)
        """ cosine of all elements in array """

        sum = staticmethod(torch.sum)
        """ sum elements in array """

        max = staticmethod(torch.max)
        """ max element in array """

        stack = staticmethod(torch.stack)
        """ stack multiple arrays """

        @staticmethod
        def transpose(arr, axes=None):
            """transpose array by flipping two dimensions"""
            if axes is None:
                axes = tuple(range(len(arr.shape) - 1, -1, -1))
            return arr.permute(*axes)

        squeeze = staticmethod(torch.squeeze)
        """ remove dim-1 dimensions """

        broadcast_arrays = staticmethod(torch.broadcast_tensors)
        """ broadcast arrays """

        broadcast_to = staticmethod(torch.broadcast_to)
        """ broadcast array into shape """

        reshape = staticmethod(torch.reshape)
        """ reshape array into given shape """

        bmm = staticmethod(torch.bmm)
        """ batch matrix multiply two arrays """

        @staticmethod
        def is_array(arr):
            """check if an object is an array"""
            # is this a reasonable implemenation?
            return isinstance(arr, numpy.ndarray) or torch.is_tensor(arr)

        def array(self, arr, dtype=None):
            """create an array from an array-like sequence"""
            if dtype is None:
                dtype = torch.get_default_dtype()
            if torch.is_tensor(arr):
                return arr.clone().to(device="cpu", dtype=dtype)
            return torch.tensor(arr, device="cpu", dtype=dtype)

        # constructors
        ones = staticmethod(torch.ones)
        """ create an array filled with ones """

        zeros = staticmethod(torch.zeros)
        """ create an array filled with zeros """

        def linspace(self, start, stop, num=50, endpoint=True):
            """create a linearly spaced array between two points"""
            delta = (stop - start) / float(num - float(endpoint))
            if not delta:
                return self.array([start] * num)
            return torch.arange(start, stop + 0.5 * float(endpoint) * delta, delta)

        arange = staticmethod(torch.arange)
        """ create a range of values """

        pad = staticmethod(torch.nn.functional.pad)  # type: ignore

        fftfreq = staticmethod(numpy.fft.fftfreq)

        fft = staticmethod(torch.fft)  # type: ignore

        divide = staticmethod(torch.div)

        exp = staticmethod(torch.exp)
        # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
        # The same warning applies here.
        # <3 <3 <3 <3

        def numpy(self, arr):
            """convert the array to numpy array"""
            if torch.is_tensor(arr):
                return arr.numpy()
            else:
                return numpy.asarray(arr)

    # Torch Cuda Backend
    if TORCH_CUDA_AVAILABLE:

        class TorchCudaBackend(TorchBackend):
            """Torch Cuda Backend"""

            def ones(self, shape):
                """create an array filled with ones"""
                return torch.ones(shape, device="cuda")

            def zeros(self, shape):
                """create an array filled with zeros"""
                return torch.zeros(shape, device="cuda")

            def array(self, arr, dtype=None):
                """create an array from an array-like sequence"""
                if dtype is None:
                    dtype = torch.get_default_dtype()
                if torch.is_tensor(arr):
                    return arr.clone().to(device="cuda", dtype=dtype)
                return torch.tensor(arr, device="cuda", dtype=dtype)

            # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            # The same warning applies here.
            def numpy(self, arr):
                """convert the array to numpy array"""
                if torch.is_tensor(arr):
                    return arr.cpu().numpy()
                else:
                    return numpy.asarray(arr)

            def linspace(self, start, stop, num=50, endpoint=True):
                """convert a linearly spaced interval of values"""
                delta = (stop - start) / float(num - float(endpoint))
                if not delta:
                    return self.array([start] * num)
                return torch.arange(
                    start, stop + 0.5 * float(endpoint) * delta, delta, device="cuda"
                )


## Default Backend
# this backend object will be used for all array/tensor operations.
# the backend is changed by changing the class of the backend
# using the "set_backend" function. This "monkeypatch" will replace all the methods
# of the backend object by the methods supplied by the new class.
backend = NumpyBackend()


## Set backend
def set_backend(name: str):
    """Set the backend for the FDTD simulations

    This function monkeypatches the backend object by changing its class.
    This way, all methods of the backend object will be replaced.

    Args:
        name: name of the backend. Allowed backend names:
            - ``numpy`` (defaults to float64 arrays)
            - ``numpy.float16``
            - ``numpy.float32``
            - ``numpy.float64``
            - ``numpy.float128``
            - ``torch`` (defaults to float64 tensors)
            - ``torch.float16``
            - ``torch.float32``
            - ``torch.float64``
            - ``torch.cuda`` (defaults to float64 tensors)
            - ``torch.cuda.float16``
            - ``torch.cuda.float32``
            - ``torch.cuda.float64``

    """
    # perform checks
    if name.startswith("torch") and not TORCH_AVAILABLE:
        raise RuntimeError("Torch backend is not available. Is PyTorch installed?")
    if name.startswith("torch.cuda") and not TORCH_CUDA_AVAILABLE:
        raise RuntimeError(
            "Torch cuda backend is not available.\n"
            "Do you have a GPU on your computer?\n"
            "Is PyTorch with cuda support installed?"
        )

    if name.count(".") == 0:
        dtype, device = "float64", "cpu"
    elif name.count(".") == 1:
        name, dtype = name.split(".")
        if dtype == "cuda":
            device, dtype = "cuda", "float64"
        else:
            device = "cpu"
    elif name.count(".") == 2:
        name, device, dtype = name.split(".")
    else:
        raise ValueError(f"Unknown backend '{name}'")

    if name == "numpy":
        if device == "cpu":
            backend.__class__ = NumpyBackend
            backend.float = getattr(numpy, dtype)
        elif device == "cuda":
            raise ValueError(
                "Device 'cuda' not available for numpy backend. Use 'torch' backend in stead."
            )
        else:
            raise ValueError(
                "Unknown device '{device}'. Available devices: 'cpu', 'cuda'"
            )
    elif name == "torch":
        if device == "cpu":
            backend.__class__ = TorchBackend
            backend.float = getattr(torch, dtype)
        elif device == "cuda":
            backend.__class__ = TorchCudaBackend
            backend.float = getattr(torch, dtype)
        else:
            raise ValueError(
                "Unknown device '{device}'. Available devices: 'cpu', 'cuda'"
            )
    else:
        raise ValueError(
            "Unknown backend '{name}'. Available backends: 'numpy', 'torch'"
        )



from math import sqrt

from . import constants as const

"""
Optional reduced unit conversion functions for user use.
Mainly to make it explicit where conversions happen.

Unit system described by the thesis
"Novel architectures for brain-inspired photonic computers",
https://www.photonics.intec.ugent.be/download/phd_259.pdf

Chapters 4.1.2 and 4.1.6.

> "In SI units, the relative magnitude difference between the fields is related
by the electromagnetic impedance of free space, which for the current choice
of simulation units equals 1 per definition".

FIXME: DC: find and add md notes on scaling

"""

# Might perhaps be worth putting a note in the readme about default units / suggested unit systems?
# (done, see next PR)
# Also, if /fdtd/ gets the ability to dump to VTK, one would hate to
# have to scale everything for physical results in paraview. Will put a scaling option there.
# Is it reasonable to have a global flag for scaling? probably not
#
# don't know how severe noise is with fp32, whether it's worth the extra hassle for most people


def simE_to_worldE(input):
    return input / sqrt(const.eps0)


def worldE_to_simE(input):
    return sqrt(const.eps0) * input


def simH_to_worldH(input):
    return input / sqrt(const.mu0)


def worldH_to_simH(input):
    return sqrt(const.mu0) * input



# reverse literate programming?

from math import sin, cos, sqrt, log, exp

import numpy as np

# For Hanning window pulses
def hanning(f, t, n):
    return (1 / 2) * (1 - cos(f * t / n)) * (sin(f * t))


"""

Recipes for Z measurements usually specify gaussian or gaussian derivative input pulses.

Presumably because unlike a unit step or simply gated sine,
they're smooth, introducing minimal numerical noise, while retaining broadband frequency components.

I don't know what advantages the gaussian derivative provides over the gaussian.

These gaussian pulses are normalized to have a peak value of 1.0.

For derivation, see
https://github.com/0xDBFB7/fdtd_PCB_extensions/blob/master/normalizedgaussian.nb.pdf

http://www.cse.yorku.ca/~kosta/CompVis_Notes/fourier_transform_Gaussian.pdf
http://www.sci.utah.edu/~gerig/CS7960-S2010/handouts/04%20Gaussian%20derivatives.pdf

"""

fwhm_constant = 2.0 * sqrt(2.0 * log(2))


def normalized_gaussian_pulse(x, fwhm, center=0.0):
    # '''
    # x - input value, usually time
    # fwhm - pulse Full Width Half Maximum
    # Center - just for convenience, used to move pulse time around
    # '''
    # apparently FWHM of t is properly called "full duration half maximum",
    # but I've never heard that used

    # TODO: flag on input type and use backend?

    sigma = fwhm / fwhm_constant
    return np.exp(-(((x - center) ** 2.0) / (2.0 * (sigma ** 2.0))))


def normalized_gaussian_derivative_pulse(x, fwhm, center=0.0):
    sigma = fwhm / fwhm_constant
    return (
        exp((1.0 / 2.0) - ((x - center) ** 2.0) / (2.0 * sigma ** 2.0)) * (x - center)
    ) / sigma



""" visualization methods for the fdtd Grid.

This module supplies visualization methods for the FDTD Grid. They are
imported by the Grid class and hence are available as Grid methods.

"""

## Imports
import os

# plotting
import matplotlib.pyplot as plt
import matplotlib.patches as ptc
from matplotlib.colors import LogNorm

# 3rd party
from tqdm import tqdm
from numpy import log10, where
from scipy.signal import hilbert  # TODO: Write hilbert function to replace using scipy

# relative
from .backend import backend as bd


# 2D visualization function


def visualize(
    grid,
    x=None,
    y=None,
    z=None,
    cmap="Blues",
    pbcolor="C3",
    pmlcolor=(0, 0, 0, 0.1),
    objcolor=(1, 0, 0, 0.1),
    srccolor="C0",
    detcolor="C2",
    norm="linear",
    show=False,  # default False to allow animate to be true
    animate=False,  # True to see frame by frame states of grid while running simulation
    index=None,  # index for each frame of animation (visualize fn runs in a loop, loop variable is passed as index)
    save=False,  # True to save frames (requires parameters index, folder)
    folder=None,  # folder path to save frames
):
    """visualize a projection of the grid and the optical energy inside the grid

    Args:
        x: the x-value to make the yz-projection (leave None if using different projection)
        y: the y-value to make the zx-projection (leave None if using different projection)
        z: the z-value to make the xy-projection (leave None if using different projection)
        cmap: the colormap to visualize the energy in the grid
        pbcolor: the color to visualize the periodic boundaries
        pmlcolor: the color to visualize the PML
        objcolor: the color to visualize the objects in the grid
        srccolor: the color to visualize the sources in the grid
        detcolor: the color to visualize the detectors in the grid
        norm: how to normalize the grid_energy color map ('linear' or 'log').
        show: call pyplot.show() at the end of the function
        animate: see frame by frame state of grid during simulation
        index: index for each frame of animation (typically a loop variable is passed)
        save: save frames in a folder
        folder: path to folder to save frames
    """
    if norm not in ("linear", "lin", "log"):
        raise ValueError("Color map normalization should be 'linear' or 'log'.")
    # imports (placed here to circumvent circular imports)
    from .sources import PointSource, LineSource, PlaneSource
    from .boundaries import _PeriodicBoundaryX, _PeriodicBoundaryY, _PeriodicBoundaryZ
    from .boundaries import (
        _PMLXlow,
        _PMLXhigh,
        _PMLYlow,
        _PMLYhigh,
        _PMLZlow,
        _PMLZhigh,
    )

    if animate:  # pause for 0.1s, clear plot
        plt.pause(0.02)
        plt.clf()
        plt.ion()  # ionteration on for animation effect

    # validate x, y and z
    if x is not None:
        if not isinstance(x, int):
            raise ValueError("the `x`-location supplied should be a single integer")
        if y is not None or z is not None:
            raise ValueError(
                "if an `x`-location is supplied, one should not supply a `y` or a `z`-location!"
            )
    elif y is not None:
        if not isinstance(y, int):
            raise ValueError("the `y`-location supplied should be a single integer")
        if z is not None or x is not None:
            raise ValueError(
                "if a `y`-location is supplied, one should not supply a `z` or a `x`-location!"
            )
    elif z is not None:
        if not isinstance(z, int):
            raise ValueError("the `z`-location supplied should be a single integer")
        if x is not None or y is not None:
            raise ValueError(
                "if a `z`-location is supplied, one should not supply a `x` or a `y`-location!"
            )
    else:
        raise ValueError(
            "at least one projection plane (x, y or z) should be supplied to visualize the grid!"
        )

    # just to create the right legend entries:
    plt.plot([], lw=7, color=objcolor, label="Objects")
    plt.plot([], lw=7, color=pmlcolor, label="PML")
    plt.plot([], lw=3, color=pbcolor, label="Periodic Boundaries")
    plt.plot([], lw=3, color=srccolor, label="Sources")
    plt.plot([], lw=3, color=detcolor, label="Detectors")

    # Grid energy
    grid_energy = bd.sum(grid.E ** 2 + grid.H ** 2, -1)
    if x is not None:
        assert grid.Ny > 1 and grid.Nz > 1
        xlabel, ylabel = "y", "z"
        Nx, Ny = grid.Ny, grid.Nz
        pbx, pby = _PeriodicBoundaryY, _PeriodicBoundaryZ
        pmlxl, pmlxh, pmlyl, pmlyh = _PMLYlow, _PMLYhigh, _PMLZlow, _PMLZhigh
        grid_energy = grid_energy[x, :, :]
    elif y is not None:
        assert grid.Nx > 1 and grid.Nz > 1
        xlabel, ylabel = "z", "x"
        Nx, Ny = grid.Nz, grid.Nx
        pbx, pby = _PeriodicBoundaryZ, _PeriodicBoundaryX
        pmlxl, pmlxh, pmlyl, pmlyh = _PMLZlow, _PMLZhigh, _PMLXlow, _PMLXhigh
        grid_energy = grid_energy[:, y, :].T
    elif z is not None:
        assert grid.Nx > 1 and grid.Ny > 1
        xlabel, ylabel = "x", "y"
        Nx, Ny = grid.Nx, grid.Ny
        pbx, pby = _PeriodicBoundaryX, _PeriodicBoundaryY
        pmlxl, pmlxh, pmlyl, pmlyh = _PMLXlow, _PMLXhigh, _PMLYlow, _PMLYhigh
        grid_energy = grid_energy[:, :, z]
    else:
        raise ValueError("Visualization only works for 2D grids")

    for source in grid.sources:
        if isinstance(source, LineSource):
            if x is not None:
                _x = [source.y[0], source.y[-1]]
                _y = [source.z[0], source.z[-1]]
            elif y is not None:
                _x = [source.z[0], source.z[-1]]
                _y = [source.x[0], source.x[-1]]
            elif z is not None:
                _x = [source.x[0], source.x[-1]]
                _y = [source.y[0], source.y[-1]]
            plt.plot(_y, _x, lw=3, color=srccolor)
        elif isinstance(source, PointSource):
            if x is not None:
                _x = source.y
                _y = source.z
            elif y is not None:
                _x = source.z
                _y = source.y
            elif z is not None:
                _x = source.x
                _y = source.y
            plt.plot(_y - 0.5, _x - 0.5, lw=3, marker="o", color=srccolor)
            grid_energy[_x, _y] = 0  # do not visualize energy at location of source
        elif isinstance(source, PlaneSource):
            if x is not None:
                _x = (
                    source.y
                    if source.y.stop > source.y.start + 1
                    else slice(source.y.start, source.y.start)
                )
                _y = (
                    source.z
                    if source.z.stop > source.z.start + 1
                    else slice(source.z.start, source.z.start)
                )
            elif y is not None:
                _x = (
                    source.z
                    if source.z.stop > source.z.start + 1
                    else slice(source.z.start, source.z.start)
                )
                _y = (
                    source.x
                    if source.x.stop > source.x.start + 1
                    else slice(source.x.start, source.x.start)
                )
            elif z is not None:
                _x = (
                    source.x
                    if source.x.stop > source.x.start + 1
                    else slice(source.x.start, source.x.start)
                )
                _y = (
                    source.y
                    if source.y.stop > source.y.start + 1
                    else slice(source.y.start, source.y.start)
                )
            patch = ptc.Rectangle(
                xy=(_y.start - 0.5, _x.start - 0.5),
                width=_y.stop - _y.start,
                height=_x.stop - _x.start,
                linewidth=0,
                edgecolor="none",
                facecolor=srccolor,
            )
            plt.gca().add_patch(patch)

    # Detector
    for detector in grid.detectors:
        if x is not None:
            _x = [detector.y[0], detector.y[-1]]
            _y = [detector.z[0], detector.z[-1]]
        elif y is not None:
            _x = [detector.z[0], detector.z[-1]]
            _y = [detector.x[0], detector.x[-1]]
        elif z is not None:
            _x = [detector.x[0], detector.x[-1]]
            _y = [detector.y[0], detector.y[-1]]

        if detector.__class__.__name__ == "BlockDetector":
            # BlockDetector
            plt.plot(
                [_y[0], _y[1], _y[1], _y[0], _y[0]],
                [_x[0], _x[0], _x[1], _x[1], _x[0]],
                lw=3,
                color=detcolor,
            )
        else:
            # LineDetector
            plt.plot(_y, _x, lw=3, color=detcolor)

    # Boundaries
    for boundary in grid.boundaries:
        if isinstance(boundary, pbx):
            _x = [-0.5, -0.5, float("nan"), Nx - 0.5, Nx - 0.5]
            _y = [-0.5, Ny - 0.5, float("nan"), -0.5, Ny - 0.5]
            plt.plot(_y, _x, color=pbcolor, linewidth=3)
        elif isinstance(boundary, pby):
            _x = [-0.5, Nx - 0.5, float("nan"), -0.5, Nx - 0.5]
            _y = [-0.5, -0.5, float("nan"), Ny - 0.5, Ny - 0.5]
            plt.plot(_y, _x, color=pbcolor, linewidth=3)
        elif isinstance(boundary, pmlyl):
            patch = ptc.Rectangle(
                xy=(-0.5, -0.5),
                width=boundary.thickness,
                height=Nx,
                linewidth=0,
                edgecolor="none",
                facecolor=pmlcolor,
            )
            plt.gca().add_patch(patch)
        elif isinstance(boundary, pmlxl):
            patch = ptc.Rectangle(
                xy=(-0.5, -0.5),
                width=Ny,
                height=boundary.thickness,
                linewidth=0,
                edgecolor="none",
                facecolor=pmlcolor,
            )
            plt.gca().add_patch(patch)
        elif isinstance(boundary, pmlyh):
            patch = ptc.Rectangle(
                xy=(Ny - 0.5 - boundary.thickness, -0.5),
                width=boundary.thickness,
                height=Nx,
                linewidth=0,
                edgecolor="none",
                facecolor=pmlcolor,
            )
            plt.gca().add_patch(patch)
        elif isinstance(boundary, pmlxh):
            patch = ptc.Rectangle(
                xy=(-0.5, Nx - boundary.thickness - 0.5),
                width=Ny,
                height=boundary.thickness,
                linewidth=0,
                edgecolor="none",
                facecolor=pmlcolor,
            )
            plt.gca().add_patch(patch)

    for obj in grid.objects:
        if x is not None:
            _x = (obj.y.start, obj.y.stop)
            _y = (obj.z.start, obj.z.stop)
        elif y is not None:
            _x = (obj.z.start, obj.z.stop)
            _y = (obj.x.start, obj.x.stop)
        elif z is not None:
            _x = (obj.x.start, obj.x.stop)
            _y = (obj.y.start, obj.y.stop)

        patch = ptc.Rectangle(
            xy=(min(_y) - 0.5, min(_x) - 0.5),
            width=max(_y) - min(_y),
            height=max(_x) - min(_x),
            linewidth=0,
            edgecolor="none",
            facecolor=objcolor,
        )
        plt.gca().add_patch(patch)

    # visualize the energy in the grid
    cmap_norm = None
    if norm == "log":
        cmap_norm = LogNorm(vmin=1e-4, vmax=grid_energy.max() + 1e-4)
    plt.imshow(bd.numpy(grid_energy), cmap=cmap, interpolation="sinc", norm=cmap_norm)

    # finalize the plot
    plt.ylabel(xlabel)
    plt.xlabel(ylabel)
    plt.ylim(Nx, -1)
    plt.xlim(-1, Ny)
    plt.figlegend()
    plt.tight_layout()

    # save frame (require folder path and index)
    if save:
        plt.savefig(os.path.join(folder, f"file{str(index).zfill(4)}.png"))

    # show if not animating
    if show:
        plt.show()


def dB_map_2D(block_det=None, choose_axis=2, interpolation="spline16"):
    """
    Displays detector readings from an 'fdtd.BlockDetector' in a decibel map spanning a 2D slice region inside the BlockDetector.
    Compatible with continuous sources (not pulse).
    Currently, only x-y 2D plot slices are accepted.

    Parameter:-
        block_det (numpy array): 5 axes numpy array (timestep, row, column, height, {x, y, z} parameter) created by 'fdtd.BlockDetector'.
        (optional) choose_axis (int): Choose between {0, 1, 2} to display {x, y, z} data. Default 2 (-> z).
        (optional) interpolation (string): Preferred 'matplotlib.pyplot.imshow' interpolation. Default "spline16".
    """
    if block_det is None:
        raise ValueError(
            "Function 'dBmap' requires a detector_readings object as parameter."
        )
    if len(block_det.shape) != 5:  # BlockDetector readings object have 5 axes
        raise ValueError(
            "Function 'dBmap' requires object of readings recorded by 'fdtd.BlockDetector'."
        )

    # TODO: convert all 2D slices (y-z, x-z plots) into x-y plot data structure

    plt.ioff()
    plt.close()
    a = []  # array to store wave intensities
    for i in tqdm(range(len(block_det[0]))):
        a.append([])
        for j in range(len(block_det[0][0])):
            temp = [x[i][j][0][choose_axis] for x in block_det]
            a[i].append(max(temp) - min(temp))

    peakVal, minVal = max(map(max, a)), min(map(min, a))
    print(
        "Peak at:",
        [
            [[i, j] for j, y in enumerate(x) if y == peakVal]
            for i, x in enumerate(a)
            if peakVal in x
        ],
    )
    a = 10 * log10([[y / minVal for y in x] for x in a])

    plt.title("dB map of Electrical waves in detector region")
    plt.imshow(a, cmap="inferno", interpolation=interpolation)
    cbar = plt.colorbar()
    cbar.ax.set_ylabel("dB scale", rotation=270)
    plt.show()


def plot_detection(detector_dict=None, specific_plot=None):
    """
    1. Plots intensity readings on array of 'fdtd.LineDetector' as a function of timestep.
    2. Plots time of arrival of pulse at different LineDetector in array.
    Compatible with pulse sources.

    Args:
        detector_dict (dictionary): Dictionary of detector readings, as created by 'fdtd.Grid.save_data()'.
        (optional) specific_plot (string): Plot for a specific axis data. Choose from {"Ex", "Ey", "Ez", "Hx", "Hy", "Hz"}.
    """
    if detector_dict is None:
        raise Exception(
            "Function plotDetection() requires a dictionary of detector readings as 'detector_dict' parameter."
        )
    detectorElement = 0  # cell to consider in each detectors
    maxArray = {}
    plt.ioff()
    plt.close()

    for detector in detector_dict:
        if len(detector_dict[detector].shape) != 3:
            print("Detector '{}' not LineDetector; dumped.".format(detector))
            continue
        if specific_plot is not None:
            if detector[-2] != specific_plot[0]:
                continue
        if detector[-2] == "E":
            plt.figure(0, figsize=(15, 15))
        elif detector[-2] == "H":
            plt.figure(1, figsize=(15, 15))
        for dimension in range(len(detector_dict[detector][0][0])):
            if specific_plot is not None:
                if ["x", "y", "z"].index(specific_plot[1]) != dimension:
                    continue
            # if specific_plot, subplot on 1x1, else subplot on 2x2
            plt.subplot(
                2 - int(specific_plot is not None),
                2 - int(specific_plot is not None),
                dimension + 1 if specific_plot is None else 1,
            )
            hilbertPlot = abs(
                hilbert([x[0][dimension] for x in detector_dict[detector]])
            )
            plt.plot(hilbertPlot, label=detector)
            plt.title(detector[-2] + "(" + ["x", "y", "z"][dimension] + ")")
            if detector[-2] not in maxArray:
                maxArray[detector[-2]] = {}
            if str(dimension) not in maxArray[detector[-2]]:
                maxArray[detector[-2]][str(dimension)] = []
            maxArray[detector[-2]][str(dimension)].append(
                [detector, where(hilbertPlot == max(hilbertPlot))[0][0]]
            )

    # Loop same as above, only to add axes labels
    for i in range(2):
        if specific_plot is not None:
            if ["E", "H"][i] != specific_plot[0]:
                continue
        plt.figure(i)
        for dimension in range(len(detector_dict[detector][0][0])):
            if specific_plot is not None:
                if ["x", "y", "z"].index(specific_plot[1]) != dimension:
                    continue
            plt.subplot(
                2 - int(specific_plot is not None),
                2 - int(specific_plot is not None),
                dimension + 1 if specific_plot is None else 1,
            )
            plt.xlabel("Time steps")
            plt.ylabel("Magnitude")
        plt.suptitle("Intensity profile")
    plt.legend()
    plt.show()

    for item in maxArray:
        plt.figure(figsize=(15, 15))
        for dimension in maxArray[item]:
            arrival = bd.numpy(maxArray[item][dimension])
            plt.plot(
                [int(x) for x in arrival.T[1]],
                arrival.T[0],
                label=["x", "y", "z"][int(dimension)],
            )
        plt.title(item)
        plt.xlabel("Time of arrival (time steps)")
        plt.legend()
        plt.suptitle("Time-of-arrival plot")
    plt.show()


#
# def dump_to_vtk(pcb, filename, iteration, Ex_dump=False, Ey_dump=False, Ez_dump=False, Emag_dump=True, objects_dump=True, ports_dump=True):
#     '''
#     Extension is automatically chosen, you don't need to supply it
#
#     thanks
#     https://pyscience.wordpress.com/2014/09/06/numpy-to-vtk-converting-your-numpy-arrays-to-vtk-arrays-and-files/
#     https://bitbucket.org/pauloh/pyevtk/src/default/src/hl.py
#
#     Paraview needs a threshold operation to view the objects correctly.
#
#
#     argument scaling_factor=None,True(epsilon and hbar), or Float, to accomodate reduced units
#     or maybe
#     '''



""" The FDTD Grid

The grid is the core of the FDTD Library. It is where everything comes
together and where the biggest part of the calculations are done.

"""

## Imports

# standard library
import os
from os import path, makedirs, chdir, remove
from subprocess import check_call, CalledProcessError
from glob import glob
from datetime import datetime

# 3rd party
from tqdm import tqdm
from numpy import savez

# typing
from .typing_ import Tuple, Number, Tensorlike

# relative
from .backend import backend as bd
from . import constants as const

## Functions
def curl_E(E: Tensorlike) -> Tensorlike:
    """Transforms an E-type field into an H-type field by performing a curl
    operation

    Args:
        E: Electric field to take the curl of (E-type field located on the
           edges of the grid cell [integer gridpoints])

    Returns:
        The curl of E (H-type field located on the faces of the grid [half-integer grid points])
    """
    curl = bd.zeros(E.shape)

    curl[:, :-1, :, 0] += E[:, 1:, :, 2] - E[:, :-1, :, 2]
    curl[:, :, :-1, 0] -= E[:, :, 1:, 1] - E[:, :, :-1, 1]

    curl[:, :, :-1, 1] += E[:, :, 1:, 0] - E[:, :, :-1, 0]
    curl[:-1, :, :, 1] -= E[1:, :, :, 2] - E[:-1, :, :, 2]

    curl[:-1, :, :, 2] += E[1:, :, :, 1] - E[:-1, :, :, 1]
    curl[:, :-1, :, 2] -= E[:, 1:, :, 0] - E[:, :-1, :, 0]

    return curl


def curl_H(H: Tensorlike) -> Tensorlike:
    """Transforms an H-type field into an E-type field by performing a curl
    operation

    Args:
        H: Magnetic field to take the curl of (H-type field located on half-integer grid points)

    Returns:
        The curl of H (E-type field located on the edges of the grid [integer grid points])

    """
    curl = bd.zeros(H.shape)

    curl[:, 1:, :, 0] += H[:, 1:, :, 2] - H[:, :-1, :, 2]
    curl[:, :, 1:, 0] -= H[:, :, 1:, 1] - H[:, :, :-1, 1]

    curl[:, :, 1:, 1] += H[:, :, 1:, 0] - H[:, :, :-1, 0]
    curl[1:, :, :, 1] -= H[1:, :, :, 2] - H[:-1, :, :, 2]

    curl[1:, :, :, 2] += H[1:, :, :, 1] - H[:-1, :, :, 1]
    curl[:, 1:, :, 2] -= H[:, 1:, :, 0] - H[:, :-1, :, 0]

    return curl


## FDTD Grid Class
class Grid:
    """The FDTD Grid

    The grid is the core of the FDTD Library. It is where everything comes
    together and where the biggest part of the calculations are done.

    """

    from .visualization import visualize

    def __init__(
        self,
        shape: Tuple[Number, Number, Number],
        grid_spacing: float = 155e-9,
        permittivity: float = 1.0,
        permeability: float = 1.0,
        courant_number: float = None,
    ):
        """
        Args:
            shape: shape of the FDTD grid.
            grid_spacing: distance between the grid cells.
            permittivity: the relative permittivity of the background.
            permeability: the relative permeability of the background.
            courant_number: the courant number of the FDTD simulation.
                Defaults to the inverse of the square root of the number of
                dimensions > 1 (optimal value). The timestep of the simulation
                will be derived from this number using the CFL-condition.
        """
        # save the grid spacing
        self.grid_spacing = float(grid_spacing)

        # save grid shape as integers
        self.Nx, self.Ny, self.Nz = self._handle_tuple(shape)

        # dimension of the simulation:
        self.D = int(self.Nx > 1) + int(self.Ny > 1) + int(self.Nz > 1)

        # courant number of the simulation (optimal value)
        max_courant_number = float(self.D) ** (-0.5)
        if courant_number is None:
            # slight stability factor added
            self.courant_number = 0.99 * max_courant_number
        elif courant_number > max_courant_number:
            raise ValueError(
                f"courant_number {courant_number} too high for "
                f"a {self.D}D simulation"
            )
        else:
            self.courant_number = float(courant_number)

        # timestep of the simulation
        self.time_step = self.courant_number * self.grid_spacing / const.c

        # save electric and magnetic field
        self.E = bd.zeros((self.Nx, self.Ny, self.Nz, 3))
        self.H = bd.zeros((self.Nx, self.Ny, self.Nz, 3))

        # save the inverse of the relative permittiviy and the relative permeability
        # these tensors can be anisotropic!

        if bd.is_array(permittivity) and len(permittivity.shape) == 3:
            permittivity = permittivity[:, :, :, None]
        self.inverse_permittivity = bd.ones((self.Nx, self.Ny, self.Nz, 3)) / bd.array(
            permittivity, dtype=bd.float
        )

        if bd.is_array(permeability) and len(permeability.shape) == 3:
            permeability = permeability[:, :, :, None]
        self.inverse_permeability = bd.ones((self.Nx, self.Ny, self.Nz, 3)) / bd.array(
            permeability, dtype=bd.float
        )

        # save current time index
        self.time_steps_passed = 0

        # dictionary containing the sources:
        self.sources = []

        # dictionary containing the boundaries
        self.boundaries = []

        # dictionary containing the detectors
        self.detectors = []

        # dictionary containing the objects in the grid
        self.objects = []

        # folder path to store the simulation
        self.folder = None

    def _handle_distance(self, distance: Number) -> int:
        """transform a distance to an integer number of gridpoints"""
        if not isinstance(distance, int):
            return int(float(distance) / self.grid_spacing + 0.5)
        return distance

    def _handle_time(self, time: Number) -> int:
        """transform a time value to an integer number of timesteps"""
        if not isinstance(time, int):
            return int(float(time) / self.time_step + 0.5)
        return time

    def _handle_tuple(
        self, shape: Tuple[Number, Number, Number]
    ) -> Tuple[int, int, int]:
        """validate the grid shape and transform to a length-3 tuple of ints"""
        if len(shape) != 3:
            raise ValueError(
                f"invalid grid shape {shape}\n"
                f"grid shape should be a 3D tuple containing floats or ints"
            )
        x, y, z = shape
        x = self._handle_distance(x)
        y = self._handle_distance(y)
        z = self._handle_distance(z)
        return x, y, z

    def _handle_slice(self, s: slice) -> slice:
        """validate the slice and transform possibly float values to ints"""
        start = (
            s.start
            if not isinstance(s.start, float)
            else self._handle_distance(s.start)
        )
        stop = (
            s.stop if not isinstance(s.stop, float) else self._handle_distance(s.stop)
        )
        step = (
            s.step if not isinstance(s.step, float) else self._handle_distance(s.step)
        )
        return slice(start, stop, step)

    def _handle_single_key(self, key):
        """transform a single index key to a slice or list"""
        try:
            len(key)
            return [self._handle_distance(k) for k in key]
        except TypeError:
            if isinstance(key, slice):
                return self._handle_slice(key)
            else:
                return [self._handle_distance(key)]
        return key

    @property
    def x(self) -> int:
        """get the number of grid cells in the x-direction"""
        return self.Nx * self.grid_spacing

    @property
    def y(self) -> int:
        """get the number of grid cells in the y-direction"""
        return self.Ny * self.grid_spacing

    @property
    def z(self) -> int:
        """get the number of grid cells in the y-direction"""
        return self.Nz * self.grid_spacing

    @property
    def shape(self) -> Tuple[int, int, int]:
        """get the shape of the FDTD grid"""
        return (self.Nx, self.Ny, self.Nz)

    @property
    def time_passed(self) -> float:
        """get the total time passed"""
        return self.time_steps_passed * self.time_step

    def run(self, total_time: Number, progress_bar: bool = True):
        """run an FDTD simulation.

        Args:
            total_time: the total time for the simulation to run.
            progress_bar: choose to show a progress bar during
                simulation

        """
        if isinstance(total_time, float):
            total_time /= self.time_step
        time = range(0, int(total_time), 1)
        if progress_bar:
            time = tqdm(time)
        for _ in time:
            self.step()

    def step(self):
        """do a single FDTD step by first updating the electric field and then
        updating the magnetic field
        """
        self.update_E()
        self.update_H()
        self.time_steps_passed += 1

    def update_E(self):
        """update the electric field by using the curl of the magnetic field"""

        # update boundaries: step 1
        for boundary in self.boundaries:
            boundary.update_phi_E()

        curl = curl_H(self.H)
        self.E += self.courant_number * self.inverse_permittivity * curl

        # update objects
        for obj in self.objects:
            obj.update_E(curl)

        # update boundaries: step 2
        for boundary in self.boundaries:
            boundary.update_E()

        # add sources to grid:
        for src in self.sources:
            src.update_E()

        # detect electric field
        for det in self.detectors:
            det.detect_E()

    def update_H(self):
        """update the magnetic field by using the curl of the electric field"""

        # update boundaries: step 1
        for boundary in self.boundaries:
            boundary.update_phi_H()

        curl = curl_E(self.E)
        self.H -= self.courant_number * self.inverse_permeability * curl

        # update objects
        for obj in self.objects:
            obj.update_H(curl)

        # update boundaries: step 2
        for boundary in self.boundaries:
            boundary.update_H()

        # add sources to grid:
        for src in self.sources:
            src.update_H()

        # detect electric field
        for det in self.detectors:
            det.detect_H()

    def reset(self):
        """reset the grid by setting all fields to zero"""
        self.H *= 0.0
        self.E *= 0.0
        self.time_steps_passed *= 0

    def add_source(self, name, source):
        """add a source to the grid"""
        source._register_grid(self)
        self.sources[name] = source

    def add_boundary(self, name, boundary):
        """add a boundary to the grid"""
        boundary._register_grid(self)
        self.boundaries[name] = boundary

    def add_detector(self, name, detector):
        """add a detector to the grid"""
        detector._register_grid(self)
        self.detectors[name] = detector

    def add_object(self, name, obj):
        """add an object to the grid"""
        obj._register_grid(self)
        self.objects[name] = obj

    def __setitem__(self, key, attr):
        if not isinstance(key, tuple):
            x, y, z = key, slice(None), slice(None)
        elif len(key) == 1:
            x, y, z = key[0], slice(None), slice(None)
        elif len(key) == 2:
            x, y, z = key[0], key[1], slice(None)
        elif len(key) == 3:
            x, y, z = key
        else:
            raise KeyError("maximum number of indices for the grid is 3")

        attr._register_grid(
            grid=self,
            x=self._handle_single_key(x),
            y=self._handle_single_key(y),
            z=self._handle_single_key(z),
        )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(shape=({self.Nx},{self.Ny},{self.Nz}), "
            f"grid_spacing={self.grid_spacing:.2e}, courant_number={self.courant_number:.2f})"
        )

    def __str__(self):
        """string representation of the grid

        lists all the components and their locations in the grid.
        """
        s = repr(self) + "\n"
        if self.sources:
            s = s + "\nsources:\n"
            for src in self.sources:
                s += str(src)
        if self.detectors:
            s = s + "\ndetectors:\n"
            for det in self.detectors:
                s += str(det)
        if self.boundaries:
            s = s + "\nboundaries:\n"
            for bnd in self.boundaries:
                s += str(bnd)
        if self.objects:
            s = s + "\nobjects:\n"
            for obj in self.objects:
                s += str(obj)
        return s

    def save_simulation(self, sim_name=None):
        """
        Creates a folder and initializes environment to store simulation or related details.
        saveSimulation() needs to be run before running any function that stores data (generate_video(), save_data()).

        Parameters:-
            (optional) sim_name (string): Preferred name for simulation
        """
        makedirs("fdtd_output", exist_ok=True)  # Output master folder declaration
        # making full_sim_name with timestamp
        full_sim_name = (
            str(datetime.now().year)
            + "-"
            + str(datetime.now().month)
            + "-"
            + str(datetime.now().day)
            + "-"
            + str(datetime.now().hour)
            + "-"
            + str(datetime.now().minute)
            + "-"
            + str(datetime.now().second)
        )
        # Simulation name (optional)
        if sim_name is not None:
            full_sim_name = full_sim_name + " (" + sim_name + ")"
        folder = "fdtd_output_" + full_sim_name
        # storing folder path for saving simulation
        self.folder = os.path.abspath(path.join("fdtd_output", folder))
        # storing timestamp title for self.generate_video
        self.full_sim_name = full_sim_name
        makedirs(self.folder, exist_ok=True)
        return self.folder

    def generate_video(self, delete_frames=False):
        """Compiles frames into a video

        These framed should be saved through ``fdtd.Grid.visualize(save=True)`` while having ``fdtd.Grid.save_simulation()`` enabled.

        Args:
            delete_frames (optional, bool): delete stored frames after conversion to video.

        Returns:
            the filename of the generated video.

        Note:
            this function requires ``ffmpeg`` to be available in your path.
        """
        if self.folder is None:
            raise Exception(
                "Save location not initialized. Please read about 'fdtd.Grid.saveSimulation()' or try running 'grid.saveSimulation()'."
            )
        cwd = path.abspath(os.getcwd())
        chdir(self.folder)
        try:
            check_call(
                [
                    "ffmpeg",
                    "-y",
                    "-framerate",
                    "8",
                    "-i",
                    "file%04d.png",
                    "-r",
                    "30",
                    "-pix_fmt",
                    "yuv420p",
                    "fdtd_sim_video_" + self.full_sim_name + ".mp4",
                ]
            )
        except (FileNotFoundError, CalledProcessError):
            raise CalledProcessError(
                "Error when calling ffmpeg. Is ffmpeg installed and available in your path?"
            )
        if delete_frames:  # delete frames
            for file_name in glob("*.png"):
                remove(file_name)
        video_path = path.abspath(
            path.join(self.folder, f"fdtd_sim_video_{self.full_sim_name}.mp4")
        )
        chdir(cwd)
        return video_path

    def save_data(self):
        """
        Saves readings from all detectors in the grid into a numpy zip file. Each detector is stored in separate arrays. Electric and magnetic field field readings of each detector are also stored separately with suffix " (E)" and " (H)" (Example: ['detector0 (E)', 'detector0 (H)']). Therefore, the numpy zip file contains arrays twice the number of detectors.
        REQUIRES 'fdtd.Grid.save_simulation()' to be run before this function.

        Parameters: None
        """
        if self.folder is None:
            raise Exception(
                "Save location not initialized. Please read about 'fdtd.Grid.saveSimulation()' or try running 'grid.saveSimulation()'."
            )
        dic = {}
        for detector in self.detectors:
            dic[detector.name + " (E)"] = [x for x in detector.detector_values()["E"]]
            dic[detector.name + " (H)"] = [x for x in detector.detector_values()["H"]]
        savez(path.join(self.folder, "detector_readings"), **dic)
